from typing import List

import torch

from mppq.ir.base.opdef import Variable
from mppq.logger import nest
from mppq.quant import (
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
)
from mppq.quantization.observer.base import OBSERVER_TABLE, BaseTensorObserver
from mppq.utils.qfunction.linear import minmax_to_scale_offset


@OBSERVER_TABLE.register("minmax")
class TorchMinMaxObserver(BaseTensorObserver):
    """TorchMinMaxObserver collects min and max value of given tensor."""

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        self._min_val_collector: List[torch.Tensor] = []
        self._max_val_collector: List[torch.Tensor] = []
        self._logger = nest(self.__class__.__name__)

    @torch.no_grad()
    def observe(self, value: torch.Tensor):
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return
        assert isinstance(
            value, torch.Tensor
        ), "TorchMinMaxObserver can only deal with torch Tensor values"
        assert (
            value.numel() > 0
        ), f"You are observing an empty tensor({self._watch_on.name})."

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            self._min_val_collector.append(value.min().reshape(shape=[1]))
            self._max_val_collector.append(value.max().reshape(shape=[1]))
        elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            channel_axis = self._quant_cfg.channel_axis
            channelwise_view = value.transpose(dim0=0, dim1=channel_axis).unsqueeze(-1)
            channelwise_view = torch.flatten(channelwise_view, start_dim=1)
            self._min_val_collector.append(
                torch.min(channelwise_view, dim=1, keepdim=True)[0]
            )
            self._max_val_collector.append(
                torch.max(channelwise_view, dim=1, keepdim=True)[0]
            )
        else:
            raise TypeError(
                "Min-max Observer only work with per-tensor or per-channel "
                "quantize policy."
            )

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if len(self._max_val_collector) == 0:
            raise ValueError(
                "Can not render quantization config yet, "
                "Observer data collator is empty. "
                "Invoke observe() function before render config."
            )
        device = self._max_val_collector[-1].device

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            min_val = torch.min(torch.cat(self._min_val_collector, dim=0)).cpu().item()
            max_val = torch.max(torch.cat(self._max_val_collector, dim=0)).cpu().item()
            scale, offset = minmax_to_scale_offset(
                min_val=min_val, max_val=max_val, config=self._quant_cfg
            )
            self._logger.debug(
                f"{self._watch_on.name} observes min_val={min_val}, max_val={max_val}, "
                f"scale={scale}, zero point={offset}"
            )
            self._quant_cfg.scale = torch.tensor(
                [scale], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.offset = torch.tensor(
                [offset], dtype=torch.float32, device=device
            ).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
        elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            min_vals = (
                torch.min(
                    torch.cat(self._min_val_collector, dim=-1), dim=-1, keepdim=False
                )[0]
                .cpu()
                .numpy()
            )
            max_vals = (
                torch.max(
                    torch.cat(self._max_val_collector, dim=-1), dim=-1, keepdim=False
                )[0]
                .cpu()
                .numpy()
            )
            assert len(min_vals) == len(
                max_vals
            ), "Min values and max values should at same length."
            scales, offsets = [], []
            for min_val, max_val in zip(min_vals, max_vals):
                scale, offset = minmax_to_scale_offset(
                    min_val=min_val, max_val=max_val, config=self._quant_cfg
                )
                scales.append(scale)
                offsets.append(offset)
            self._logger.debug(
                f"{self._watch_on.name} observes min_val={min_vals}, "
                f"max_val={max_vals}, scale={scales}, zero point={offsets}"
            )
            # scale, offset here only deployed on cpu
            # we will move them towards target device through RunnableGraph
            self._quant_cfg.scale = torch.tensor(
                scales, dtype=torch.float32, device=device
            )
            self._quant_cfg.offset = torch.tensor(
                offsets, dtype=torch.float32, device=device
            )
            self._quant_cfg.state = QuantizationStates.ACTIVATED
        else:
            raise TypeError(
                "Min-max Observer only work with per-tensor or per-channel "
                "quantize policy."
            )



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

A register that can register any types or objects.

Example:

.. code-block:: python

    FOO = Registry("FOO")

    class Foo: ...

    # register directly
    FOO.register(name="foo_snake_style")(Foo)

    # use as a decorator
    @FOO.register()
    class Bar: ...

    # update as a dict
    FOO.update(Bar=NewBar, Foo=Foo)

    print(FOO)
    # ┌─────────────────┐
    # │ Register: FOO   │
    # ├─────────────────┤
    # │ foo_snake_style │
    # │ Bar             │
    # └─────────────────┘
"""

import inspect
from typing import (
    Callable,
    Dict,
    Generic,
    Iterator,
    Literal,
    Mapping,
    Optional,
    TypeVar,
    cast,
)

from tabulate import tabulate

T = TypeVar("T", bound=Callable)
F = TypeVar("F", bound=Callable)


class Registry(Generic[T]):
    """A simple registry object to hold objects from others

    Samples::

        FOO = Registry("FOO")

        @FOO.register()
        def foo(): ...

        print(FOO)
        # ┌───────────────┐
        # │ Register: FOO │
        # ├───────────────┤
        # │ foo           │
        # └───────────────┘
    """

    def __init__(
        self,
        name: Optional[str] = None,
        base: Optional[T] = None,
        parent: Optional["Registry[T]"] = None,
        name_style: Optional[Literal["snake_case", "PascalCase"]] = None,
    ) -> None:
        self._bucks: Dict[str, T] = {}
        self._configs: Dict[str, inspect.Signature] = {}
        self._name = name or "<Registry>"
        self._base = base
        self._parent = parent
        if parent is not None:
            self._name = f"{parent.name}.{self.name}"
        self._style = name_style

    @property
    def name(self) -> str:
        """Return the name of the registry."""
        return self._name

    def _legal_name(self, name: str) -> str:
        if not self._style:
            return name
        words = [""]
        for a, b in zip(list(name), list(name.lower())):
            if a != b:
                words.append("")
            words[-1] += b
        refine_words = [""]
        for w in words:
            if len(w) == 1:
                refine_words[-1] += w
            else:
                refine_words.append(w)
        if refine_words[0] == "":
            refine_words = refine_words[1:]
        if self._style == "snake_case":
            return "_".join(refine_words).strip("_")
        elif self._style == "PascalCase":
            return "".join(w.capitalize() for w in refine_words)
        else:
            return name

    def register(self, name: Optional[str] = None):
        """A decorator to register an object.

        Args:
            name (str, optional): The name of the object. If not provided, the name
                of the function of class will be used after transform to lowercase.
        """

        def wrapper(func: F) -> F:
            if not callable(func):
                raise TypeError(
                    "the object to be registered must be a function or Callable,"
                    f" got {type(func)}"
                )
            _name = name or self._legal_name(func.__name__)
            self._bucks[_name] = cast(T, func)
            self._configs[_name] = inspect.signature(func)
            if not inspect.isfunction(func):
                if self._base is not None:
                    assert isinstance(func, type) and isinstance(self._base, type)
                    if not issubclass(func, self._base):
                        raise TypeError(
                            f"the registered object {func} must be the subclass "
                            f"of {self._base}, but its mro is {func.__mro__}"
                        )
            if self._parent is not None:
                self._parent.register(name)(func)
            # forward the signature of the original function
            return cast(F, func)

        return wrapper

    def update(self, obj: Optional[Mapping[str, T]] = None, **kwargs: T) -> None:
        """Update the registry with a mapping or keyword arguments.

        Args:
            obj (Mapping[str, T]): A mapping of name to object.
            **kwargs (T): Keyword arguments of object.
        """
        if obj is not None and not isinstance(obj, Mapping):
            raise TypeError(f"obj must be a mapping, got {type(obj)}")
        if obj is not None:
            for name, func in obj.items():
                self.register(name)(func)
        for name, func in kwargs.items():
            self.register(name)(func)

    def get(self, name: str) -> Optional[T]:
        """Get a registered object by its name."""
        if name in self._bucks:
            functor = self._bucks[name]
            return functor

    def get_config(self, name: str):
        """Get the configuration of an object"""
        return self._configs.get(name)

    def query_name(self, cls: T) -> str:
        """Query registered name of a class"""
        for key, value in self._bucks.items():
            if value is cls:
                return key
        raise KeyError(f"{cls.__name__} is not registered in {self._name}")

    def __getitem__(self, name: str) -> T:
        """Get a registered object by its name."""
        obj = self.get(name)
        if obj is None:
            raise KeyError(f"{name} is not registered in {self._name}")
        return obj

    def __iter__(self) -> Iterator[str]:
        """Return an Iterator for all registered functions"""
        yield from self._bucks.keys()

    def __contains__(self, name: str | T | None) -> bool:
        """Check if a function is registered"""
        if name is None:
            return False
        if isinstance(name, str):
            return name in self._bucks
        return name in self._bucks.values()

    def __repr__(self) -> str:
        title = [f"Register: {self.name}", "Config"]
        members = []
        for i in sorted(self._bucks.keys()):
            members.append([i, self._configs[i]])
        return tabulate(members, title, "simple_grid", maxcolwidths=[None, 150])



from abc import ABCMeta, abstractmethod
from typing import Generator, Iterable, List, Optional, Self, Sequence, Type

from mppq.executor.base import BaseGraphExecutor
from mppq.ir.base.graph import BaseGraph
from mppq.logger import error, info
from mppq.register import Registry


class QuantizationOptimizationPass(metaclass=ABCMeta):
    r"""QuantizationOptimizationPass is a basic building block of PPQ
    quantization logic.

    PPQ is designed as a Multi pass Compiler of quantization network.
        where pass here refers to a traversal through the entire network.

    This class is an abstract base class of all customized passes.
    Quantizer will build an optimization pipeline later to quantize and optimize
    your network.
    """

    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name or self.__class__.__name__

    @abstractmethod
    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        **kwargs,
    ) -> Optional[BaseGraph]:
        """Apply an optimization algorithm to a given graph."""
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"QuantizationOptimizationPass[{self.name}]"


class QuantizationOptimizationPipeline(QuantizationOptimizationPass):
    """QuantizationOptimizationPipeline is a sorted set PPQ Optimization
    passes.

    PPQ is designed as a Multi pass Compiler of quantization network.
        where pass here refers to a traversal through the entire network.

    Quantizer is going to calling optimization pass from pipeline one by one to
        eventually finish network quantization procedure
    """

    def __init__(self, passes: Sequence[QuantizationOptimizationPass]):
        super().__init__(name="Quantization Optimization Pipeline")
        self._pipeline: List[QuantizationOptimizationPass] = []
        for optim in passes:
            self.append(optim)

    def __len__(self) -> int:
        return len([_ for _ in self.__iter__()])

    def __contains__(self, obj: QuantizationOptimizationPass) -> bool:
        return obj in self._pipeline

    def __iter__(self) -> Generator[QuantizationOptimizationPass, None, None]:
        for i in self._pipeline:
            if isinstance(i, QuantizationOptimizationPipeline):
                # support nested pipeline
                yield from i.__iter__()
            else:
                yield i

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Optional[Iterable] = None,
        executor: Optional[BaseGraphExecutor] = None,
        verbose: bool = True,
        **kwargs,
    ):
        for i, optim in enumerate(self):
            assert isinstance(optim, QuantizationOptimizationPass)
            if verbose:
                info(f"[{i + 1:02d}/{len(self):02d} {optim.name} Running ...")

            if not isinstance(graph, BaseGraph):
                raise TypeError(
                    "parameter 1 should be an instance of PPQ BaseGraph when "
                    f"calling optim pass, however {type(graph)} was given."
                )
            try:
                ret = optim.optimize(
                    graph=graph,
                    dataloader=dataloader,
                    executor=executor,
                    verbose=verbose,
                    **kwargs,
                )
            except Exception:
                error(f"An error occurred while running {optim.name}.")
                raise
            finally:
                if verbose:
                    info(f"{optim.name} Finished.")
            if isinstance(ret, BaseGraph):
                graph = ret
        return graph

    def append(self, optim: QuantizationOptimizationPass) -> Self:
        """Add an optimization pass to the end of pipeline."""
        if not isinstance(optim, QuantizationOptimizationPass):
            error("Pass should be a subclass of QuantizationOptimizationPass")
            error(f"Unexpected pass type: {type(optim)}")
            raise TypeError
        self._pipeline.append(optim)
        return self

    def prepend(self, optim: QuantizationOptimizationPass) -> Self:
        """Add an optimization pass to the beginning of pipeline."""
        if not isinstance(optim, QuantizationOptimizationPass):
            error("Pass should be a subclass of QuantizationOptimizationPass")
            error(f"Unexpected pass type: {type(optim)}")
            raise TypeError
        self._pipeline = [optim] + self._pipeline
        return self

    def __repr__(self) -> str:
        report = ""
        for optimization_pass in self._pipeline:
            report += str(optimization_pass) + "\n"
        return report


OPTIM_ALGORITHMS: Registry[Type[QuantizationOptimizationPass]] = Registry(
    "OPTIM_ALGORITHMS"
)



from mppq.quantization.measure.cosine import (
    numpy_cosine_similarity,
    torch_cosine_similarity,
    torch_cosine_similarity_as_loss,
)
from mppq.quantization.measure.norm import torch_mean_square_error, torch_snr_error
from mppq.quantization.measure.statistic import torch_KL_divergence

__all__ = [
    "numpy_cosine_similarity",
    "torch_cosine_similarity",
    "torch_cosine_similarity_as_loss",
    "torch_mean_square_error",
    "torch_snr_error",
    "torch_KL_divergence",
]



from typing import Any, List, Optional

import torch

from mppq.data import (
    DataType,
    convert_any_to_python_primary_type,
    convert_any_to_tensor,
)
from mppq.ir.base.command import (
    GraphCommand,
    GraphCommandType,
    ReplaceOperationCommand,
    ReplaceVariableCommand,
    TruncateGraphCommand,
)
from mppq.ir.base.graph import Operation, Variable
from mppq.ir.base.processor import GraphCommandProcessor
from mppq.ir.search import SearchableGraph
from mppq.logger import error, warning


class GraphReplacer(GraphCommandProcessor):
    r"""Graph Replacer offers a bunch of graph editing functions that helps
    replacing operation or variable in your graph.
    """

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.REPLACE_OP:
            assert isinstance(
                command, ReplaceOperationCommand
            ), "Use ReplaceOperationCommand instead of GraphCommand"
            return self.replace_op(command.op_name, command.replace_to)
        if command.command_type == GraphCommandType.REPLACE_VAR:
            assert isinstance(
                command, ReplaceVariableCommand
            ), "Use ReplaceOperationCommand instead of GraphCommand"
            return self.replace_var(command.op_name, command.replace_to)
        if command.command_type == GraphCommandType.REPLACE_BATCHNORM_TO_CONV:
            return self.replace_batchnorm_to_conv()
        if command.command_type == GraphCommandType.REPLACE_BATCHNORM_TO_SCALE:
            return self.replace_batchnorm_to_scale()

    def replace_op(self, op_name: str, replace_to: Operation):
        if op_name not in self._graph.operations:
            error(f"Operation {op_name} is not in current graph")
            raise KeyError
        operation = self._graph.operations[op_name]

        replace_to.inputs.clear()
        replace_to.inputs.extend(operation.inputs)
        for input_var in operation.inputs:
            dest_idx = input_var.dest_ops.index(operation)
            input_var.dest_ops[dest_idx] = replace_to

        replace_to.outputs.clear()
        replace_to.outputs.extend(operation.outputs)
        for output_var in operation.outputs:
            output_var.source_op = replace_to

        replace_to.parameters.clear()
        replace_to.parameters.extend(operation.parameters)

        self._graph.operations[op_name] = replace_to

    def replace_var(self, var_name: str, replace_to: Variable):
        if var_name not in self._graph.variables:
            error(f"Variable {var_name} is not in current graph")
            raise KeyError
        variable = self._graph.variables[var_name]

        replace_to.dest_ops.clear()
        replace_to.dest_ops.extend(variable.dest_ops)
        for dest_op in replace_to.dest_ops:
            dest_idx = dest_op.inputs.index(variable)
            dest_op.inputs[dest_idx] = replace_to

        replace_to.source_op = variable.source_op
        if variable.source_op is not None:
            source_idx = variable.source_op.outputs.index(variable)
            variable.source_op.outputs[source_idx] = replace_to

        self._graph.variables[var_name] = replace_to
        if var_name in self._graph.inputs:
            self._graph.inputs[var_name] = replace_to
        if var_name in self._graph.outputs:
            self._graph.outputs[var_name] = replace_to

    @property
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            GraphCommandType.REPLACE_OP,
            GraphCommandType.REPLACE_VAR,
            GraphCommandType.REPLACE_BATCHNORM_TO_CONV,
            GraphCommandType.REPLACE_BATCHNORM_TO_SCALE,
        ]

    def replace_batchnorm_to_conv(self, dimension: int = 2):
        """Replace Batchnorm to 1D Convolution."""
        for op in self.graph.operations.values():
            if op.type == "BatchNormalization":
                warning(
                    f"Isolated BatchNormalization({op.name}) was detected, "
                    f"PPQ will replace it to 1*1 Convolution({dimension}D)."
                )

                assert (
                    len(op.parameters) == 4
                ), "BatchNorm should have 4 parameters, namely alpha, beta, mean, var"
                alpha = op.parameters[0].value
                beta = op.parameters[1].value
                mean = op.parameters[2].value
                var = op.parameters[3].value
                epsilon = op.attributes.get("epsilon", 1e-5)

                with torch.no_grad():
                    w = alpha / torch.sqrt(var + epsilon)
                    w = w.reshape([-1, 1] + [1] * dimension)
                    b = alpha * (-mean) / torch.sqrt(var + epsilon) + beta

                op.type = "Conv"
                op.attributes.clear()
                op.attributes["kernel_shape"] = [1] * dimension
                op.attributes["strides"] = [1] * dimension
                op.attributes["dilations"] = [1] * dimension
                op.attributes["pads"] = [0, 0] * dimension
                op.attributes["group"] = w.numel()

                # remove last 2 variable, make conv has exact 3 input
                self.graph.remove_variable(op.inputs[-1])
                self.graph.remove_variable(op.inputs[-1])

                with torch.no_grad():
                    op.inputs[1].value = w
                    op.inputs[2].value = b

    def replace_batchnorm_to_scale(self, dimension: int = 4):
        """Replace Batchnorm to Mul + Add.

        By default this function created a 4d mul + add corresponding to NCHW layout.
        """
        graph = self.graph
        for op in [_ for _ in self.graph.operations.values()]:

            if op.type == "BatchNormalization":
                warning(
                    f"Isolated BatchNormalization({op.name}) was detected, "
                    f"PPQ will replace it to Mul + Add({dimension}D)."
                )

                assert (
                    len(op.parameters) == 4
                ), "BatchNorm should have 4 parameters, namely alpha, beta, mean, var"
                alpha = op.parameters[0].value
                beta = op.parameters[1].value
                mean = op.parameters[2].value
                var = op.parameters[3].value
                epsilon = op.attributes.get("epsilon", 1e-5)

                with torch.no_grad():
                    multiplier = alpha / torch.sqrt(var + epsilon)
                    bias = (-mean) * multiplier + beta

                for var in [_ for _ in op.parameters]:
                    graph.remove_variable(var)
                graph.create_variable(
                    value=multiplier, is_parameter=True, dest_ops=[op]
                )
                op.type = "Mul"
                op.attributes.clear()

                add = graph.create_operation(op_type="Add")
                graph.insert_op_after(op0=add, op1=op)
                graph.create_variable(value=bias, is_parameter=True, dest_ops=[add])

                if dimension > 1:
                    op.parameters[0].value = op.parameters[0].value.reshape(
                        [1, -1] + [1] * (dimension - 2)
                    )
                    add.parameters[0].value = add.parameters[0].value.reshape(
                        [1, -1] + [1] * (dimension - 2)
                    )


class GraphFormatter(GraphCommandProcessor):
    """Graph Formatter offers a bunch of graph editing functions that helps
    modifying your graph.
    """

    @property
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            GraphCommandType.FORMAT_CLIP,
            GraphCommandType.FORMAT_PAD,
            GraphCommandType.FORMAT_GATHER,
            GraphCommandType.FORMAT_CAST,
            GraphCommandType.FORMAT_INT64_CONSTANT,
            GraphCommandType.DELETE_ISOLATED,
            GraphCommandType.FORMAT_PARAMETERS,
            GraphCommandType.FORMAT_CONSTANT_INPUT,
            GraphCommandType.FORMAT_SLICE,
            GraphCommandType.TRUNCATE_ON_VAR,
            GraphCommandType.FORMAT_RESIZE,
            GraphCommandType.REMOVE_IDENTITY,
            GraphCommandType.CONVERT_TO_TENSOR,
        ]

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.FORMAT_CLIP:
            return self.format_clip()
        if command.command_type == GraphCommandType.FORMAT_PAD:
            return self.format_pad()
        if command.command_type == GraphCommandType.FORMAT_GATHER:
            return self.format_gather()
        if command.command_type == GraphCommandType.FORMAT_CAST:
            return self.format_cast()
        if command.command_type == GraphCommandType.DELETE_ISOLATED:
            return self.delete_isolated()
        if command.command_type == GraphCommandType.FORMAT_INT64_CONSTANT:
            return self.format_int64_constant()
        if command.command_type == GraphCommandType.FORMAT_PARAMETERS:
            return self.format_parameter()
        if command.command_type == GraphCommandType.FORMAT_CONSTANT_INPUT:
            return self.remove_constant_input()
        if command.command_type == GraphCommandType.FORMAT_SLICE:
            return self.format_slice()
        if command.command_type == GraphCommandType.FORMAT_RESIZE:
            return self.format_resize()
        if command.command_type == GraphCommandType.TRUNCATE_ON_VAR:
            assert isinstance(
                command, TruncateGraphCommand
            ), "Use TruncateGraphCommand here."
            return self.truncate_on_var(command.var, command.mark_as_output)
        if command.command_type == GraphCommandType.REMOVE_IDENTITY:
            return self.remove_identity()
        if command.command_type == GraphCommandType.CONVERT_TO_TENSOR:
            return self.convert_to_tensor()

    def format_slice(self) -> None:
        """
        Slice: opset1 格式跟其他的不太一样，这个 pass 将 opset1 的 slice 强行转换为 opset 11
        """
        interested_ops = []
        for operation in self.graph.operations.values():
            if operation.type == "Slice":
                if "starts" in operation.attributes:
                    assert (
                        "starts" in operation.attributes
                        and "ends" in operation.attributes
                    ), (
                        f"Invalid Slice Operation Format, Slice operation is expected "
                        "to have axes, starts and ends attributes with opset 1, "
                        f"however your operation {operation.name}, do not have "
                        "completed attributes"
                    )
                    interested_ops.append(operation)

        for slice_op in interested_ops:
            assert isinstance(slice_op, Operation)
            axes = slice_op.attributes.get("axes", None)
            starts = slice_op.attributes["starts"]
            ends = slice_op.attributes["ends"]
            if axes is None:
                axes = [_ for _ in range(starts)]

            slice_op.attributes.pop("starts")
            slice_op.attributes.pop("ends")
            if "axes" in slice_op.attributes:
                slice_op.attributes.pop("axes")
            self.__add_constant_input(slice_op, convert_any_to_tensor(starts))
            self.__add_constant_input(slice_op, convert_any_to_tensor(ends))
            self.__add_constant_input(slice_op, convert_any_to_tensor(axes))

    def format_pad(self) -> None:
        """
        对于不同的模型格式, pad 算子将有两种不同的输入格式：
        for different models, possibly Pad op has the following input formats
            1. pads 参数由第二个输入变量给出
               pads parameter is given by the second input variable
            2. pads 参数被放置于 operation.attribute 中
               pads parameter is set in attribute
        此函数统一 pad 算子行为：所有 pad 算子的 pads 参数均由第二个输入给出
        """
        for op in self.graph.operations.values():
            if op.type == "Pad" and "pads" in op.attributes:
                self.graph.create_variable(
                    value=torch.tensor(op.attributes["pads"]),
                    is_parameter=True,
                    dest_ops=[op],
                )
                op.attributes.clear()

    def format_resize(self) -> None:
        """
        升级 opset 10 的 resize 到 opset 11
        """
        for op in self.graph.operations.values():
            if op.type == "Resize" and len(op.inputs) == 2:
                # 创建一个空的 variable, 这个 variable 没有值与 source_op, 也不是 parameter
                # 这种行为 Onnx 是允许的，这种变量用以传递空值，起到占位符的作用
                self.graph.create_variable(
                    value=None, is_parameter=False, dest_ops=[op]
                )
                op.inputs[1], op.inputs[2] = op.inputs[2], op.inputs[1]

    def format_clip(self) -> None:
        r"""对于不同的模型格式, clip 算子将有两种不同的输入格式：
            1. min, max 参数由 第二、第三个输入变量给出
            2. min, max 参数由 attribute 给出
        此函数统一 clip 算子行为：所有 clip 算子的 min, max 参数第二第三个变量给出
        针对可能存在的 min, max 为空的情况，将其直接置为 2 << 30（保证处理后非空）
        当 min, max 参数由 第二、第三个输入变量给出时，其中一个为空时直接返回 ValueError
        """

        interested_ops = []
        for _, operation in self.graph.operations.items():
            if operation.type == "Clip" and (
                "min" in operation.attributes or "max" in operation.attributes
            ):
                interested_ops.append(operation)
        for op in interested_ops:
            assert isinstance(op, Operation)
            min_value = op.attributes.get("min", -2 << 30)
            max_value = op.attributes.get("max", +2 << 30)
            min_var = Variable(
                name=op.name + "_min", value=min_value, is_parameter=True, dest_ops=[op]
            )
            max_var = Variable(
                name=op.name + "_max", value=max_value, is_parameter=True, dest_ops=[op]
            )
            self.graph.append_variable(min_var)
            self.graph.append_variable(max_var)
            op.inputs.append(min_var)
            op.inputs.append(max_var)
            if "min" in op.attributes:
                op.attributes.pop("min")
            if "max" in op.attributes:
                op.attributes.pop("max")

    def format_gather(self) -> None:
        """gather op 的参数 index 可能由 input variable 给出 但 index
        参数不可以被量化，同时后端运算需要其作为Python 原生类型 因此将其转移到 gather op 的属性上。

        gather op 的参数 axis 可能不存在，此时强制植入 0 作为 axis 参数
        """
        interested_ops = []
        for _, operation in self.graph.operations.items():
            if operation.type == "Gather":
                interested_ops.append(operation)
        for operation in interested_ops:
            assert isinstance(operation, Operation)
            if len(operation.inputs) == 2:
                index_op = operation.inputs[1].source_op
                if index_op is not None and index_op.type == "Constant":
                    index = index_op.attributes["value"]
                    index_value = convert_any_to_python_primary_type(index)
                    self.__delete_constant_input(operation, 1)
                    operation.attributes["gather_index"] = index_value
            if "axis" not in operation.attributes:
                operation.attributes["axis"] = 0

            if "indices" in operation.attributes:
                operation.attributes["gather_index"] = operation.attributes["indices"]
                operation.attributes.pop("indices")

    def format_cast(self) -> None:
        """cast op 的参数 to 默认为 int，使用该函数将其封装为 ppq.core.DataType."""
        interested_ops = []
        for _, operation in self.graph.operations.items():
            assert isinstance(operation, Operation)
            if operation.type == "Cast":
                interested_ops.append(operation)
        for operation in interested_ops:
            assert isinstance(operation, Operation)
            assert "to" in operation.attributes
            operation.attributes["to"] = DataType.from_numpy(operation.attributes["to"])

    def format_int64_constant(self) -> None:
        """convert all int64 constants to int32, check if direct dtype cast is
        available 将所有 int64 的 Constant 转换为 int32 将检查所有 Constant value, 如果 value
        范围在 int32 表示范围内则执行转换。"""
        for operation in self.graph.operations.values():
            if operation.type == "Constant":
                assert "value" in operation.attributes
                value = operation.attributes["value"]

                assert isinstance(value, torch.Tensor)
                if value.dtype != torch.int64:
                    continue

                pvalue = value.tolist()
                check = [0xFFFFFFFF > v >= -0xFFFFFFFF for v in pvalue]

                if all(check):
                    value = value.int()

    def remove_constant_input(self) -> None:
        """部分部署平台不支持 Constant Op 作为算子的输入
            在这种情况下我们使用这个 pass 把它们切换成 Parameter Variable

        Some backend platform doesn't support Constant
        Op, we use this pass to replace it by forcing its value to be a
        parameter variable."""
        removing_ops = []
        for op in self.graph.operations.values():
            if op.type == "Constant":
                if op.num_of_output != 1:
                    error(f"Constant {op.name} has {op.num_of_output} outputs")
                    raise ValueError
                removing_ops.append(op)

        for const_op in removing_ops:
            assert isinstance(const_op, Operation)
            constant_value = const_op.attributes["value"]
            output_var = const_op.outputs[0]
            output_var.is_parameter = True
            output_var.value = constant_value
            self.graph.remove_operation(removing_op=const_op)

    def truncate_on_var(self, var: Variable, mark_as_output: bool):
        """从一个指定位置将图截断.

        Args:
            var (Variable): _description_
            mark_as_output (bool): _description_

        Raises:
            TypeError: _description_
            KeyError: _description_
        """
        graph = self.graph
        if not isinstance(var, Variable):
            error(f"Except Variable, however {type(var)} was given.")
            raise TypeError
        if var.name not in graph.variables:
            error(f"Can not find vairiable {var.name} in current graph")
            raise KeyError

        mark_to_delete, delete_queue, didx = set(), [], 0
        delete_queue.extend(var.dest_ops)
        while didx < len(delete_queue):
            first_op = delete_queue[didx]
            if first_op not in mark_to_delete:
                mark_to_delete.add(first_op)
                delete_queue.extend(graph.get_downstream_operations(first_op))
            didx += 1

        for operation in mark_to_delete:
            graph.remove_operation(operation)

        if mark_as_output:
            graph.mark_variable_as_graph_output(var)

        self.delete_isolated()

    def delete_isolated(self):
        """Remove Isolated Variable from Graph."""
        stop = False
        while not stop:
            blacklist = []
            # delete all operations which are not links to a valid graph output
            for op in self.graph.operations.values():
                if len(self.graph.get_downstream_operations(op)) == 0:
                    output_names = [var.name for var in op.outputs]
                    if all([name not in self.graph.outputs for name in output_names]):
                        blacklist.append(op)

            for op in blacklist:
                for var in op.outputs:
                    self.graph.remove_variable(var)
                self.graph.remove_operation(op)
            stop = len(blacklist) == 0

        stop = False
        while not stop:
            var_blacklist = set()
            # delete all variables that links to invalid operations:
            for var in self.graph.variables.values():
                # 删除无根无输出的变量
                if var.source_op is None and len(var.dest_ops) == 0:
                    var_blacklist.add(var)
                # 删除根节点不在图中的变量
                if (
                    var.source_op is not None
                    and var.source_op.name not in self.graph.operations
                ):
                    var_blacklist.add(var)
                # 删除连接到未知节点的变量
                for op in var.dest_ops:
                    if op.name not in self.graph.operations:
                        var_blacklist.add(var)
                # 删除孤立变量
                if var.source_op is None and var.name not in self.graph.inputs:
                    if len(var.dest_ops) == 0:
                        var_blacklist.add(var)
                # 没有输出的不能删...会影响算子输出顺序...
            for var in var_blacklist:
                self.graph.remove_variable(var)
            stop = len(var_blacklist) == 0

    def format_parameter(self) -> None:
        """Split parameter that has more than 1 dest ops"""
        for var in [_ for _ in self.graph.variables.values()]:
            if var.has_value:
                x = convert_any_to_tensor(var.value)
                var.value = x
            if var.is_parameter and len(var.dest_ops) > 1:
                for op in var.dest_ops:
                    created = self.graph.create_variable(
                        value=var.value.clone(), is_parameter=True
                    )
                    op.inputs[op.inputs.index(var)] = created
                    created.dest_ops.append(op)
                var.dest_ops.clear()
                self.graph.remove_variable(var)

    def __delete_constant_input(self, op: Operation, input_idx: int):
        op_name = op.name
        if op_name not in self._graph.operations:
            error(f"Operation {op_name} not in current graph.")
            raise KeyError
        operation = self._graph.operations[op_name]
        if input_idx >= len(operation.inputs):
            error(
                "Trying to delete an out-of-range input variable, "
                f"has graph been manually changed? Error at Operation {op_name}, "
                f"input_idx: {input_idx}"
            )
            raise ValueError
        input_var = operation.inputs[input_idx]
        source_op = input_var.source_op
        if source_op is None or source_op.type != "Constant":
            error(
                "Trying to delete an non-const input, "
                f"Error at Operation {op_name}, inputs[{input_idx}]"
            )
            raise ValueError
        input_var.dest_ops.pop(input_var.dest_ops.index(operation))
        operation.inputs.pop(input_idx)

        if len(input_var.dest_ops) == 0:
            self.graph.remove_operation(source_op)
            self.graph.remove_variable(input_var)

    def __add_constant_input(self, op: Operation, value: torch.Tensor):
        op_name = op.name
        if op_name not in self._graph.operations:
            error(f"Operation {op_name} not in current graph.")
            raise KeyError
        operation = self._graph.operations[op_name]
        var = Variable(
            name=f"{op_name}_{len(op.inputs) + 1}", value=value, is_parameter=True
        )
        self.graph.append_variable(var)
        var.dest_ops.append(operation)
        operation.inputs.append(var)

    def remove_identity(self):
        """Remove Identity Op from Graph."""
        removing_ops = []
        for op in self.graph.operations.values():
            if op.type == "Identity":
                removing_ops.append(op)

        for op in removing_ops:
            self.graph.remove_operation(op, keep_coherence=True)

    def convert_to_tensor(self):
        """Convert anything inside your network to torch tensor. (Usually from numpy)"""
        for var in self.graph.variables.values():
            if var.has_value:
                var.value = convert_any_to_tensor(var.value)


class GraphMerger(GraphCommandProcessor):
    """Graph Merger implements all graph fusion related functions."""

    @property
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            # add more extensions in the future
            GraphCommandType.FUSE_BN,
            GraphCommandType.FUSE_BIAS_ADD,
        ]

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.FUSE_BN:
            return self.fuse_bn()
        if command.command_type == GraphCommandType.FUSE_BIAS_ADD:
            return self.fuse_bias_add()

    def fuse_bn(self):
        """fuse batchnormalization"""
        search_engine = SearchableGraph(graph=self.graph)
        paths = search_engine.path_matching(
            sp_expr=lambda op: op.type in {"Conv", "Gemm", "ConvTranspose"},
            rp_expr=lambda x, y: False,
            ep_expr=lambda op: op.type == "BatchNormalization",
            direction="down",
        )

        for path in paths:
            path = path.tolist()
            assert len(path) == 2, "Oops seems we got something unexpected."

            computing_op, bn_op = path
            assert isinstance(computing_op, Operation) and isinstance(bn_op, Operation)

            if (
                len(self.graph.get_downstream_operations(computing_op)) != 1
                or len(self.graph.get_upstream_operations(bn_op)) != 1
            ):
                warning(
                    f"PPQ can not merge operation {computing_op.name} and {bn_op.name},"
                    " this is not suppose to happen with your network, network with "
                    "batchnorm inside might not be able to quantize and deploy."
                )
                continue

            assert (
                len(bn_op.parameters) == 4
            ), "BatchNorm should have 4 parameters, namely alpha, beta, mean, var"
            alpha = bn_op.parameters[0].value
            beta = bn_op.parameters[1].value
            mean = bn_op.parameters[2].value
            var = bn_op.parameters[3].value
            epsilon = bn_op.attributes.get("epsilon", 1e-5)

            if computing_op.num_of_parameter == 1:
                w = computing_op.parameters[0].value  # no bias.
                assert isinstance(
                    w, torch.Tensor
                ), "values of parameters are assumed as torch Tensor"
                if computing_op.type == "ConvTranspose":
                    b = torch.zeros(
                        w.shape[1] * computing_op.attributes.get("group", 1)
                    )
                elif (
                    computing_op.type == "Gemm"
                    and computing_op.attributes.get("transB", 0) == 0
                ):
                    b = torch.zeros(w.shape[1])
                else:
                    b = torch.zeros(w.shape[0])
            else:
                w, b = [var.value for var in computing_op.parameters[:2]]  # has bias.

            if computing_op.type == "Conv":

                # calculate new weight and bias
                scale = alpha / torch.sqrt(var + epsilon)
                w = w * scale.reshape([-1] + [1] * (w.ndim - 1))
                b = alpha * (b - mean) / torch.sqrt(var + epsilon) + beta

            elif computing_op.type == "Gemm":

                # calculate new weight and bias
                scale = alpha / torch.sqrt(var + epsilon)
                if computing_op.attributes.get("transB", 0):
                    w = w * scale.reshape([-1, 1])
                else:
                    w = w * scale.reshape([1, -1])
                b = alpha * (b - mean) / torch.sqrt(var + epsilon) + beta

            elif computing_op.type == "ConvTranspose":

                scale = alpha / torch.sqrt(var + epsilon)
                group = computing_op.attributes.get("group", 1)
                scale = scale.reshape([group, 1, -1] + [1] * (w.ndim - 2))
                w = w.reshape([group, -1] + list(w.shape[1:])) * scale
                w = w.reshape([w.shape[0] * w.shape[1]] + list(w.shape[2:]))
                b = alpha * (b - mean) / torch.sqrt(var + epsilon) + beta
            else:
                raise TypeError(
                    f"Unexpected op type {computing_op.type}. "
                    f"Can not merge {computing_op.name} with {bn_op.name}"
                )

            # create new op and variable
            merged_op = Operation(
                computing_op.name,
                op_type=computing_op.type,
                attributes=computing_op.attributes.copy(),
            )
            weight_var = Variable(computing_op.name + "_weight", w, True, [merged_op])
            bias_var = Variable(computing_op.name + "_bias", b, True, [merged_op])

            # replace & dirty work
            input_var = computing_op.inputs[0]
            output_var = bn_op.outputs[0]

            input_var.dest_ops.remove(computing_op)
            input_var.dest_ops.append(merged_op)

            output_var.source_op = merged_op

            # delete old operations
            computing_op.inputs.pop(0)
            bn_op.outputs.clear()
            self.graph.remove_operation(computing_op)
            self.graph.remove_operation(bn_op)

            # insert new
            self.graph.append_operation(merged_op)
            merged_op.inputs.extend([input_var, weight_var, bias_var])
            merged_op.outputs.extend([output_var])

            self.graph.append_variable(weight_var)
            self.graph.append_variable(bias_var)

    def fuse_gemm(self):
        """Fuse MatMul + add into a signal Gemm
            Single Matmul will be replaced with Gemm

        Returns:
            _type_: _description_
        """

        def _is_replaceable(op: Operation) -> bool:
            return op.inputs[0].is_parameter or op.inputs[1].is_parameter

        search_engine = SearchableGraph(graph=self.graph)
        patterns = search_engine.pattern_matching(
            patterns=["MatMul", "Add"], edges=[[0, 1]], exclusive=True
        )
        for pattern in patterns:
            matmul, add = pattern

            if not _is_replaceable(matmul):
                continue

            matmul.type = "Gemm"

            matmul_out = matmul.outputs[0]
            add_out = add.outputs[0]

            if matmul.inputs[0].is_parameter:
                temp = matmul.inputs[0]
                matmul.inputs[0] = matmul.inputs[1]
                matmul.inputs[1] = temp

            assert len(add.inputs) == 2, "Oops, seems we got some problem here."
            var1, var2 = add.inputs
            bias_var = None

            if var1.source_op == matmul and var2.is_parameter:
                bias_var = var2

            if var2.source_op == matmul and var1.is_parameter:
                bias_var = var1

            # can not find a valid bias, just skip add.
            if bias_var is None:
                continue

            if len(bias_var.value.shape) == 1:
                if bias_var.value.shape[0] == matmul.parameters[0].value.shape[-1]:
                    matmul.attributes["transB"] = 1
                    weight_val = matmul.parameters[0].value

                    matmul.parameters[0].value = weight_val.transpose(-1, -2)

                    bias_var.dest_ops.clear()
                    add.inputs.remove(bias_var)

                    # remove bias add, move bias to matmul
                    self.graph.remove_operation(add)
                    self.graph.create_link_with_op(None, matmul, bias_var)
                    self.graph.create_link_with_var(matmul_out, add_out)
                elif bias_var.value.shape[0] == matmul.parameters[0].value.shape[-2]:

                    bias_var.dest_ops.clear()
                    add.inputs.remove(bias_var)
                    # remove bias add, move bias to matmul
                    self.graph.remove_operation(add)
                    self.graph.create_link_with_op(None, matmul, bias_var)
                    self.graph.create_link_with_var(matmul_out, add_out)

        # process single gemm
        for op in self.graph.operations.values():
            if op.type == "MatMul":
                if not _is_replaceable(op):
                    continue
                op.type = "Gemm"

    def fuse_layernorm(self, exclusive_search: bool = False):  # noqa: C901
        """Fuse Layernormalization with pattern matching."""

        def _fuse(
            rm1: Operation,
            rm2: Operation,
            eps: Operation,
            scale: torch.Tensor,
            bias: Optional[torch.Tensor],
            layernorm_input_var: Variable,
            layernorm_output_var: Variable,
        ) -> None | Operation:

            if rm2.type == rm1.type == "ReduceMean":
                if "axes" not in rm1.attributes:
                    return None
                if "axes" not in rm2.attributes:
                    return None
                if rm1.attributes["axes"] != rm2.attributes["axes"]:
                    return None
                layernorm_axis = rm1.attributes["axes"]
                if isinstance(layernorm_axis, list):
                    if len(layernorm_axis) != 1:
                        return None
                    layernorm_axis = layernorm_axis[0]
                if not isinstance(layernorm_axis, int):
                    return None
            else:
                layernorm_axis = -1

            if not eps.inputs[-1].is_parameter:
                return None
            value = eps.inputs[-1].value
            value = convert_any_to_tensor(value).cpu()
            if value.numel() != 1:
                return None
            layernorm_eps = value.item()
            source_op = layernorm_output_var.source_op
            if source_op is not None:
                source_op.outputs.clear()
            layernorm = self.graph.create_operation(
                op_type="LayerNormalization",
                attributes={
                    "axis": layernorm_axis,
                    "epsilon": layernorm_eps,
                    "stash_type": 0,
                },
                inputs=[
                    layernorm_input_var,
                    self.graph.create_variable(value=scale, is_parameter=True),
                ],
                outputs=[layernorm_output_var],
            )

            if bias is not None:
                self.graph.create_link_with_op(
                    None,
                    layernorm,
                    self.graph.create_variable(value=bias, is_parameter=True),
                )
            return layernorm

        search_engine = SearchableGraph(graph=self.graph)
        fused = False

        # pattern 1:
        # y = Sub(x, ReduceMean(x))
        # Add(Mul(Div(Sqrt(Add(ReduceMean(Pow(y)))), y)))
        matches = search_engine.pattern_matching(
            patterns=[
                lambda op: True,
                lambda op: op.type in {"ReduceMean", "GlobalAveragePool"},
                "Sub",
                "Pow",
                lambda op: op.type in {"ReduceMean", "GlobalAveragePool"},
                "Add",
                "Sqrt",
                "Div",
                "Mul",
            ],
            edges=[
                [0, 1],
                [0, 2],
                [1, 2],
                [2, 3],
                [2, 7],
                [3, 4],
                [4, 5],
                [5, 6],
                [6, 7],
                [7, 8],
            ],
            exclusive=exclusive_search,
        )

        for _, rm1, sub, pow, rm2, add, sqrt, div, mul in matches:
            layernorm_ops = [rm1, sub, pow, rm2, add, sqrt, div, mul]

            layernorm_scale = mul.inputs[-1].value
            layernorm_output_var = div.outputs[0]
            layernorm_input_var = sub.inputs[0]

            # bias check
            layernorm_bias = None
            next_op = self.graph.get_downstream_operations(mul)
            if len(next_op) == 1 and (next_op[0].type == "Add"):
                bias_op = next_op[0]
                if bias_op.inputs[-1].is_parameter:
                    layernorm_bias = bias_op.inputs[-1].value
                    layernorm_output_var = bias_op.outputs[0]
                    layernorm_ops.append(bias_op)

            layernorm = _fuse(
                rm1=rm1,
                rm2=rm2,
                eps=add,
                scale=layernorm_scale,
                bias=layernorm_bias,
                layernorm_input_var=layernorm_input_var,
                layernorm_output_var=layernorm_output_var,
            )

            if layernorm is not None:
                # delete merged ops
                for op in layernorm_ops:
                    assert isinstance(op, Operation)
                    for var in op.inputs + op.outputs:
                        if var != layernorm_input_var and var != layernorm_output_var:
                            self.graph.remove_variable(var)
                    self.graph.remove_operation(op)
                    fused = True

        # pattern 2:
        matches = search_engine.pattern_matching(
            patterns=[
                lambda op: True,
                lambda op: op.type in {"ReduceMean", "GlobalAveragePool"},
                "Sub",
                "Mul",
                lambda op: op.type in {"ReduceMean", "GlobalAveragePool"},
                "Add",
                "Sqrt",
                "Reciprocal",
                "Mul",
                "Mul",
                "Sub",
                "Mul",
                "Add",
            ],
            edges=[
                [0, 1],
                [0, 2],
                [0, 11],
                [1, 2],
                [1, 9],
                [2, 3],
                [3, 4],
                [4, 5],
                [5, 6],
                [6, 7],
                [7, 8],
                [8, 9],
                [8, 11],
                [9, 10],
                [11, 12],
                [10, 12],
            ],
            exclusive=exclusive_search,
        )

        for (
            _,
            rm1,
            sub1,
            mul,
            rm2,
            add1,
            sqrt,
            recipro,
            mul2,
            mul3,
            sub2,
            mul4,
            add2,
        ) in matches:
            layernorm_ops = [
                rm1,
                sub1,
                mul,
                rm2,
                add1,
                sqrt,
                recipro,
                mul2,
                mul3,
                sub2,
                mul4,
                add2,
            ]

            # mul check
            if not mul2.inputs[-1].is_parameter:
                continue
            layernorm_scale = mul2.inputs[-1].value
            layernorm_output_var = add2.outputs[0]
            layernorm_input_var = sub1.inputs[0]
            layernorm_bias = sub2.inputs[0].value

            layernorm = _fuse(
                rm1=rm1,
                rm2=rm2,
                eps=add1,
                scale=layernorm_scale,
                bias=layernorm_bias,
                layernorm_input_var=layernorm_input_var,
                layernorm_output_var=layernorm_output_var,
            )

            if layernorm is not None:
                # delete merged ops
                for op in layernorm_ops:
                    assert isinstance(op, Operation)
                    for var in op.inputs + op.outputs:
                        if var != layernorm_input_var and var != layernorm_output_var:
                            self.graph.remove_variable(var)
                    self.graph.remove_operation(op)
                fused = True

        # final check, if no valid pattern was found, we give a warning.
        if not fused:
            warning("No valid layernorm pattern was found, check your graph again.")

    def fuse_skiplayernorm(self):
        """Fuse Add + Layernorm to SkipLayernorm, SkipLayernorm is a plugin operation
        defined by TensorRT
        """
        fused = False
        search_engine = SearchableGraph(graph=self.graph)

        matches = search_engine.pattern_matching(
            patterns=["Add", "LayerNormalization"], edges=[[0, 1]], exclusive=True
        )
        for add, layernorm in matches:
            # Skip connection can not be constant.
            if add.num_of_parameter == 0:
                input_vars = add.inputs.copy()
                output_var = add.outputs[0]
                self.graph.remove_operation(add)
                self.graph.remove_variable(output_var)

                for var in input_vars:
                    var.dest_ops.append(layernorm)
                    layernorm.inputs.append(var)

                layernorm.type = "skipLayerNormPlugin"
                fused = True
        # final check, if no valid pattern was found, we give a warning.
        if not fused:
            warning(
                "No valid Skip Layernorm pattern was found, check your graph again."
            )

    def fuse_gelu(self):
        """Fuse Gelu

        Pattern: * - Div - Erf - Add - Mul - Mul
                   |                 |
                   -------------------
        """
        fused = False
        search_engine = SearchableGraph(graph=self.graph)

        matches = search_engine.pattern_matching(
            patterns=[lambda op: True, "Div", "Erf", "Add", "Mul", "Mul"],
            edges=[[0, 1], [1, 2], [2, 3], [3, 4], [0, 4], [4, 5]],
            exclusive=True,
        )

        for _, div, erf, add, mul1, mul2 in matches:
            removing_var = []
            removing_var.extend(div.outputs)
            removing_var.extend(erf.outputs)
            removing_var.extend(add.outputs)
            removing_var.extend(mul1.outputs)

            self.graph.remove_operation(div)
            self.graph.remove_operation(erf)
            self.graph.remove_operation(add)
            self.graph.remove_operation(mul1)
            for var in removing_var:
                self.graph.remove_variable(var)

            input_vars = _.outputs.copy()
            output_vars = mul2.outputs.copy()

            self.graph.remove_operation(mul2)
            self.graph.create_operation(
                op_type="Gelu", inputs=input_vars, outputs=output_vars
            )
            assert len(input_vars) == 1, "Fusion failed, Pattern unrecognized."
            fused = True

        # final check, if no valid pattern was found, we give a warning.
        if not fused:
            warning("No valid Gelu pattern was found, check your graph again.")

    def fuse_bias_add(self):
        """
        Fuse Pattern like Conv + Add, ConvTranspose + Add, Gemm + Add
        This fusion will require a constant input as bias.
        """
        graph = self.graph
        for op in [_ for _ in graph.operations.values()]:
            if op.type in {"Conv", "ConvTranspose", "Gemm"}:
                # check if current op has only 1 downstream op
                channel_dimension = 1  # NCHW, NCHWD, NCH
                if op.type == "Gemm":
                    channel_dimension = 0
                if len(graph.get_downstream_operations(op)) == 1:
                    down = graph.get_downstream_operations(op)[0]

                    if down.type == "Add":
                        if down.num_of_parameter != 1:
                            continue
                        if op.num_of_input == 3:  # already has a bias
                            continue

                        bias = down.parameters[0]
                        if op.type not in {"Gemm"}:
                            # check if it is a bias add
                            if not bias.value.dim() == op.parameters[0].value.dim():
                                continue
                            if not bias.value.squeeze().dim() <= 1:
                                continue
                            if bias.value.shape[channel_dimension] == 1:
                                bias.value = bias.value.reshape(
                                    [-1]
                                )  # fuse add even if oc=1
                            else:
                                bias.value = (
                                    bias.value.squeeze()
                                )  # conv bias can only be 1d
                        else:
                            # Gemm bias can be any shape.
                            pass

                        # ready for fusion
                        graph.create_variable(
                            is_parameter=True, value=bias.value, dest_ops=[op]
                        )
                        graph.remove_operation(removing_op=down, keep_coherence=True)

    def fuse_selfattention(self):  # noqa: C901
        search_engine = SearchableGraph(graph=self.graph)
        matches = search_engine.pattern_matching(
            patterns=["MatMul", "Add", "Softmax", "MatMul"],
            edges=[[0, 1], [1, 2], [2, 3]],
            exclusive=False,
        )

        for m1, add, softmax, m2 in matches:
            trans_q, trans_k, trans_v, trans_o = None, None, None, None
            perm_q, perm_k, perm_v, perm_o = None, None, None, None

            # check pattern
            if m1.num_of_parameter != 0 or m2.num_of_parameter != 0:
                continue
            if m2.inputs[0].source_op != softmax:
                continue

            # source op of Q
            sq = m1.inputs[0].source_op
            if sq is not None and sq.type == "Transpose":
                trans_q = sq
                perm_q = trans_q.attributes["perm"]

            # source op of K
            sk = m1.inputs[1].source_op
            if sk is not None and sk.type == "Transpose":
                trans_k = sk
                perm_k = trans_k.attributes["perm"]

            # source op of V
            sv = m2.inputs[1].source_op
            if sv is not None and sv.type == "Transpose":
                trans_v = sv
                perm_v = trans_v.attributes["perm"]

            # output op of O
            oo = m2.outputs[0].dest_ops
            if len(oo) == 1 and oo[0].type == "Transpose":
                trans_o = oo[0]
                perm_o = trans_o.attributes["perm"]

            for op in [trans_q, trans_k, trans_v, trans_o, softmax]:
                if op is not None:
                    self.graph.remove_operation(op, keep_coherence=True)
            for var in [add.outputs[0], m1.outputs[0]]:
                self.graph.remove_variable(var)

            q_var, k_var, v_var = m1.inputs[0], m1.inputs[1], m2.inputs[1]
            mask_var, o_var = add.inputs[0], m2.outputs[0]

            for op in m1, add, m2:
                self.graph.remove_operation(op)

            op = self.graph.create_operation(
                op_type="SelfAttention",
                attributes={
                    "TransQ": perm_q,
                    "TransK": perm_k,
                    "TransV": perm_v,
                    "TransO": perm_o,
                },
                inputs=[q_var, k_var, v_var, mask_var],
                outputs=[o_var],
            )

            # remove empty transpose
            non_empty_attr = {}
            for k, v in op.attributes.values():
                if v is not None:
                    non_empty_attr[k] = v
            op.attributes = non_empty_attr

        matches = search_engine.pattern_matching(
            patterns=["MatMul", "Softmax", "MatMul"],
            edges=[[0, 1], [1, 2]],
            exclusive=False,
        )

        # self-attention with no mask
        for m1, softmax, m2 in matches:
            trans_q, trans_k, trans_v, trans_o = None, None, None, None
            perm_q, perm_k, perm_v, perm_o = None, None, None, None

            # check pattern
            if m1.num_of_parameter != 0 or m2.num_of_parameter != 0:
                continue
            if m2.inputs[0].source_op != softmax:
                continue

            # source op of Q
            sq = m1.inputs[0].source_op
            if sq is not None and sq.type == "Transpose":
                trans_q = sq
                perm_q = trans_q.attributes["perm"]

            # source op of K
            sk = m1.inputs[1].source_op
            if sk is not None and sk.type == "Transpose":
                trans_k = sk
                perm_k = trans_k.attributes["perm"]

            # source op of V
            sv = m2.inputs[1].source_op
            if sv is not None and sv.type == "Transpose":
                trans_v = sv
                perm_v = trans_v.attributes["perm"]

            # output op of O
            oo = m2.outputs[0].dest_ops
            if len(oo) == 1 and oo[0].type == "Transpose":
                trans_o = oo[0]
                perm_o = trans_o.attributes["perm"]

            for op in [trans_q, trans_k, trans_v, trans_o, softmax]:
                if op is not None:
                    self.graph.remove_operation(op, keep_coherence=True)
            for var in [m1.outputs[0]]:
                self.graph.remove_variable(var)

            q_var, k_var, v_var = m1.inputs[0], m1.inputs[1], m2.inputs[1]
            o_var = m2.outputs[0]

            for op in m1, m2:
                self.graph.remove_operation(op)

            op = self.graph.create_operation(
                op_type="SelfAttention",
                attributes={
                    "TransQ": perm_q,
                    "TransK": perm_k,
                    "TransV": perm_v,
                    "TransO": perm_o,
                },
                inputs=[q_var, k_var, v_var, self.graph.create_variable()],
                outputs=[o_var],
            )

            # remove empty transpose
            non_empty_attr = {}
            for k, v in op.attributes.values():
                if v is not None:
                    non_empty_attr[k] = v
            op.attributes = non_empty_attr



from hashlib import sha256
from itertools import product
from typing import Any, Dict, Iterator, List, Literal, Optional, Protocol, Sequence

import numpy as np
import torch
from torch import Tensor

from mppq.common import (
    COMPUTING_OP,
    DEFAULT_OPSET_DOMAIN,
    DEFAULT_OPSET_VERSION,
    LINEAR_ACTIVATIONS,
    ONNX_DOMAIN,
)
from mppq.data import DataType, convert_any_to_tensor
from mppq.logger import warning
from mppq.quant import TargetPrecision
from mppq.register import Registry
from mppq.storage import Serializable


class Opset:
    """Open Neural Network Exchange (ONNX) is an open ecosystem that empowers
    AI developers to choose the right tools as their project evolves.

    ONNX provides an open source format for AI models, both deep learning and
    traditional ML.
    It defines an extensible computation graph model, as well as definitions of
    built-in operators and standard data types.
    Currently we focus on the capabilities needed for inferencing (scoring).

    PPQ IR is built based on ONNX definition.
    """

    def __init__(
        self,
        domain: str = DEFAULT_OPSET_DOMAIN,
        version: int = DEFAULT_OPSET_VERSION,
    ) -> None:
        self.domain = domain
        self.version = version

    def is_onnx(self):
        return self.domain == ONNX_DOMAIN


class Variable(Serializable):
    r"""定义网络中的变量。通常指的是ONNX中的常量算子（Constant）和初值（Initializer）"""

    def __init__(
        self,
        name: str,
        value: Optional[Tensor] = None,
        is_parameter: bool = False,
        dest_ops: Optional[List["Operation"]] = None,
        source_op: Optional["Operation"] = None,
        shape: Optional[Sequence[int | str]] = None,
        dtype: DataType = DataType.FP32,
    ) -> None:
        super().__init__()
        self.name = name
        self.is_parameter = is_parameter
        self.dest_ops = dest_ops or []
        self._source_op = source_op
        self._value = value
        self._shape = shape
        self._dtype = dtype

    @property
    def source_op(self) -> Optional["Operation"]:
        return self._source_op

    @source_op.setter
    def source_op(self, value: Optional["Operation"]):
        self._source_op = value

    @property
    def dest_idx(self) -> List[int]:
        _dest_idx = []
        for op in self.dest_ops:
            _dest_idx.append(op.inputs.index(self))
        return _dest_idx

    @property
    def src_idx(self) -> Optional[int]:
        if self._source_op is not None:
            return self._source_op.outputs.index(self)
        return None

    @property
    def value(self) -> Tensor:
        if self._value is None:
            raise ValueError(f"Variable {self.name} has no value.")
        return self._value

    @value.setter
    def value(self, value: Tensor):
        self._value = value
        self.shape = value.shape
        self.dtype = value.dtype

    @property
    def has_value(self) -> bool:
        return self._value is not None

    @property
    def shape(self) -> Sequence[str | int]:
        if self._shape is None:
            raise ValueError(f"Variable {self.name} has no shape.")
        return self._shape

    @shape.setter
    def shape(self, value: Sequence[int | str]):
        shape = [i for i in value]
        for element in shape:
            if not isinstance(element, (int, str)):
                raise TypeError(
                    f"Shape of a variable should only contains int or str. "
                    f"however {type(element)} was given."
                )
        self._shape = shape

    @property
    def has_shape(self) -> bool:
        return self._shape is not None

    @property
    def dtype(self) -> DataType:
        if self._dtype is None:
            raise ValueError(f"Variable {self.name} has no dtype.")
        return self._dtype

    @dtype.setter
    def dtype(self, value):
        if isinstance(value, np.dtype):
            self._dtype = DataType.from_numpy(value)
        elif isinstance(value, torch.dtype):
            self._dtype = DataType.from_torch(value)
        elif isinstance(value, DataType):
            self._dtype = value
        elif isinstance(value, str):
            self._dtype = DataType.from_numpy(np.dtype(value))
        else:
            raise TypeError(f"Invalid Dtype: {value} was given.")

    def copy(self, copy_value: bool = False):
        if not copy_value or not self.has_value:
            cloned = Variable(
                name=self.name,
                value=self._value,
                is_parameter=self.is_parameter,
                shape=self._shape,
                dtype=self._dtype,
            )
            return cloned

        if not isinstance(self.value, torch.Tensor):
            warning(
                f"You are requiring to copy variable {self.name}, "
                "however its value is not an instance of torch.Tensor, "
                "ppq will automatically convert it to torch.Tensor now."
            )
            self.value = convert_any_to_tensor(self.value)
        if isinstance(self.value, torch.Tensor):
            value = self.value.clone()
        else:
            value = self.value
        return Variable(
            name=self.name,
            value=value,
            is_parameter=self.is_parameter,
            shape=self.shape,
            dtype=self.dtype,
        )

    def __hash__(self) -> int:
        return int(sha256(self.name.encode("utf-8")).hexdigest(), 16)

    def __repr__(self) -> str:
        return f"{self.name}(shape={self.shape})"


class Operation(Serializable):
    def __init__(
        self,
        name: str,
        op_type: str,
        attributes: Dict[str, Any],
        precision: TargetPrecision = TargetPrecision.UNSPECIFIED,
        inputs: Optional[List[Variable]] = None,
        outputs: Optional[List[Variable]] = None,
        opset: Optional[Opset] = None,
    ) -> None:
        Serializable.__init__(self)
        self.name = name
        self.type = op_type
        self.attributes = attributes
        self.precision = precision
        if opset is None:
            self.opset = Opset()
        else:
            self.opset = opset
        self._detail = {}
        self._input_vars = inputs or []
        self._output_vars = outputs or []

    @property
    def inputs(self) -> List[Variable]:
        return self._input_vars

    @property
    def num_of_input(self) -> int:
        return len(self.inputs)

    @property
    def outputs(self) -> List[Variable]:
        return self._output_vars

    @property
    def num_of_output(self) -> int:
        return len(self.outputs)

    @property
    def is_computing_op(self) -> bool:
        return self.type in COMPUTING_OP

    @property
    def socket(self) -> "OpSocket":
        if self.type in DEFAULT_SOCKET_TABLE:
            return DEFAULT_SOCKET_TABLE[self.type](self)
        return DEFAULT_SOCKET_CREATOR(self)

    @property
    def parameters(self) -> List[Variable]:
        return [var for var in self.inputs if var.is_parameter]

    @property
    def num_of_parameter(self) -> int:
        return len(self.parameters)

    @property
    def is_linear_activation(self) -> bool:
        return self.type in LINEAR_ACTIVATIONS

    @property
    def is_boundary(self) -> bool:
        up_ops, down_ops = [], []
        for var in self.inputs:
            up_ops.append(var.source_op)
        for var in self.outputs:
            down_ops.extend(var.dest_ops)
        return all([op is None for op in up_ops]) or len(down_ops) == 0

    def set_extension_attrib(self, attrib: str, value: Any):
        self._detail[attrib] = value

    @property
    def extension_attrib(self) -> dict:
        return self._detail

    def __hash__(self) -> int:
        return int(sha256(self.name.encode("utf-8")).hexdigest(), 16)

    def __repr__(self) -> str:
        return (
            f"{self.name}(Type: {self.type}, "
            f"Num of Input: {self.num_of_input}, "
            f"Num of Output: {self.num_of_output})"
        )

    def __deepcopy__(self, memo=None) -> "Operation":
        clone = Operation(
            name=self.name,
            op_type=self.type,
            attributes=self.attributes.copy(),
            precision=self.precision,
            opset=self.opset,
        )
        clone._detail = self._detail.copy()
        return clone


class VLink:
    r"""定义一个算子的输入输出端口之间的计算关系。"""

    def __init__(self, in_idx: int, out_idx: int) -> None:
        self.in_idx = in_idx
        self.out_idx = out_idx

    def __getitem__(self, idx: Literal[0, 1]) -> int:
        return self.in_idx if idx == 0 else self.out_idx

    def __iter__(self) -> Iterator[int]:
        yield self.in_idx
        yield self.out_idx

    def __repr__(self) -> str:
        return f"VLink({self.in_idx}, {self.out_idx})"


class OpSocket:
    r"""定义一个算子的抽象计算关系。

    in_plat属性定义了该算子每个输入端口是否可被量化。在默认情况下，所有输入均会指定为 TargetPlatform.UNSPECIFIED
    可被量化。如果一些输入不可被量化，比如Resize的scale，roi，则需要指定为 TargetPlatform.FP32 或者其他平台。
    同理，out_plat属性定义了该算子每个输出端口是否可被量化。

    links属性定义了任意一对输入输出端口之间是否有计算关系。默认情况下，任意一对输入输出端口之间均存在计算关系。
    比如 Conv(x, weight, bias) 的输出需要x，weight，bias均参与计算，它们对输出的量化参数有依赖关系。
    而 Resize(x, roi, scales) 的输出不需要参roi, scales与计算，影响输出量化参数的只有x。
    """

    def __init__(
        self,
        op: Operation,
        in_plat: Optional[List[TargetPrecision]] = None,
        out_plat: Optional[List[TargetPrecision]] = None,
        links: Optional[List[VLink]] = None,
    ) -> None:
        self.in_plat = in_plat or [TargetPrecision.UNSPECIFIED] * op.num_of_input
        self.out_plat = out_plat or [TargetPrecision.UNSPECIFIED] * op.num_of_output
        self.links = links or [
            VLink(i, j)
            for i, j in product(range(op.num_of_input), range(op.num_of_output))
        ]


def _default_socket_creator(op: Operation) -> OpSocket:
    return OpSocket(op=op)


DEFAULT_SOCKET_CREATOR = _default_socket_creator


class _OpSocketFactory(Protocol):
    def __call__(self, op: Operation) -> OpSocket:
        return DEFAULT_SOCKET_CREATOR(op)


DEFAULT_SOCKET_TABLE: Registry[_OpSocketFactory] = Registry("DEFAULT_SOCKET_TABLE")



import inspect
import math
from numbers import Number
from typing import List

import pytest
import torch

from mppq.ffi import CUDA, CUDA_COMPLIER, ENABLE_CUDA_KERNEL
from mppq.quant import RoundingPolicy
from mppq.quantization.measure.norm import torch_snr_error
from mppq.utils.round import ppq_tensor_round


# pylint: disable=missing-function-docstring
@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
def test_cuda_compile():
    with ENABLE_CUDA_KERNEL():
        dir(CUDA_COMPLIER.extension)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
@pytest.mark.parametrize(
    "size",
    [
        [1, 1, 1, 1],
        [1, 1, 1, 1],
        [5, 12, 13, 4],
        [1, 7, 15, 41],
        [50, 120, 130, 4],
        [12, 74, 15, 411],
        [50, 7, 130, 1],
        [12, 4, 15, 3],
        [5011, 7, 7, 1],
        [122552, 1, 10, 4],
        [10, 10, 124, 47],
        [19, 42, 150, 3],
    ],
)
def test_tensorwise_linear_quantize(size: List[int]):
    with ENABLE_CUDA_KERNEL():
        t = torch.rand(size=size).cuda() * 32
        s = torch.rand(size=[1]).cuda()
        o = torch.randint(low=0, high=255, size=[1]).float().cuda()

        # torch quantize
        qt = ppq_tensor_round(t / s, RoundingPolicy.ROUND_HALF_EVEN) + o
        qt = qt.clip(0, 255)
        qt = (qt - o) * s

        cuda_qt = CUDA.LinearQuantize_T(
            t, s, o, 0, 255, RoundingPolicy.ROUND_HALF_EVEN.value
        )

        diff = qt - cuda_qt
        assert diff.abs().max() == 0


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
@pytest.mark.parametrize(
    ("size", "c"),
    [
        ([1, 1, 1, 1], 1),
        ([1, 1, 1, 1], 1),
        ([5, 12, 13, 4], 1),
        ([1, 7, 15, 41], 1),
        ([50, 120, 130, 4], 1),
        ([12, 74, 15, 411], 1),
        ([50, 7, 130, 1], 1),
        ([12, 4, 15, 3], 1),
        ([5011, 7, 7, 1], 0),
        ([122552, 1, 10, 4], 0),
        ([10, 10, 124, 47], 3),
        ([19, 42, 150, 3], 3),
    ],
)
def test_channelwise_linear_quantize(size: List[int], c: int):
    with ENABLE_CUDA_KERNEL():
        nc = size[c]
        t = torch.rand(size=size).cuda() * 32
        s = torch.rand(size=[nc]).cuda()
        o = torch.randint(low=0, high=255, size=[nc]).float().cuda()

        shape = [1 if axis != c else -1 for axis in range(t.ndim)]
        s, o = s.view(shape), o.view(shape)
        qt = ppq_tensor_round(t / s, policy=RoundingPolicy.ROUND_HALF_EVEN) + o
        qt = qt.clip(0, 255)
        qt = (qt - o) * s

        cuda_qt = CUDA.LinearQuantize_C(
            t, s, o, c, 0, 255, RoundingPolicy.ROUND_HALF_EVEN.value
        )

        diff = qt - cuda_qt
        assert diff.abs().max() == 0


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
@pytest.mark.parametrize(
    "size",
    [
        [1, 1, 1, 1],
        [1, 1, 1, 1],
        [5, 12, 13, 4],
        [1, 7, 15, 41],
        [50, 120, 130, 4],
        [12, 74, 15, 411],
        [50, 7, 130, 1],
        [12, 4, 15, 3],
        [5011, 7, 7, 1],
        [122552, 1, 10, 4],
        [10, 10, 124, 47],
        [19, 42, 150, 3],
    ],
)
def test_tensorwise_linear_quantize_backward(size: List[int]):
    def ref_grad_func(
        value: torch.Tensor,
        dy: torch.Tensor,
        scale: torch.Tensor,
        offset: torch.Tensor,
    ):

        qt = ppq_tensor_round(value / scale, policy=RoundingPolicy.ROUND_HALF_EVEN)
        qt += offset
        clipped_qt = qt.clip(0, 255)

        dx = torch.where(clipped_qt != qt, torch.zeros_like(dy), dy)
        ds = torch.where(
            clipped_qt == qt,
            (((qt - offset) * scale) - value) * dy / scale,
            torch.zeros_like(dy),
        )
        ds += torch.where(qt > 255, (255 - offset) * dy, torch.zeros_like(dy))
        ds += torch.where(qt < 0, (0 - offset) * dy, torch.zeros_like(dy))
        ds = ds.sum() / math.sqrt(value.numel() * (255 - 0))
        return dx, ds

    with ENABLE_CUDA_KERNEL():
        t = torch.rand(size=size).cuda() * 50
        s = torch.rand(size=[1]).cuda()
        o = torch.randint(low=0, high=255, size=[1]).float().cuda()
        dy = torch.rand_like(t)

        cuda_grad_x, cuda_grad_s = CUDA.LinearQuantize_T_B(
            t, s, o, dy, 0, 255, RoundingPolicy.ROUND_HALF_EVEN.value
        )

        ref_grad_x, ref_grad_s = ref_grad_func(t, dy, s, o)
        diff = ref_grad_x.flatten() - cuda_grad_x.flatten()
        assert diff.abs().max() == 0

        snr = torch_snr_error(
            cuda_grad_s.reshape([1, -1]), ref_grad_s.reshape([1, -1])
        ).item()
        if snr > 1e-3:
            assert cuda_grad_s.item() <= 1


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
@pytest.mark.parametrize(
    ("size", "c"),
    [
        ([1, 1, 1, 1], 1),
        ([1, 1, 1, 1], 1),
        ([5, 12, 13, 4], 1),
        ([1, 7, 15, 41], 1),
        ([50, 120, 130, 4], 1),
        ([12, 74, 15, 411], 1),
        ([50, 7, 130, 1], 1),
        ([12, 4, 15, 3], 1),
        ([5011, 7, 7, 1], 0),
        ([122552, 1, 10, 4], 0),
        ([10, 10, 124, 47], 3),
        ([19, 42, 150, 3], 3),
    ],
)
def test_channelwise_linear_quantize_backward(size: List[int], c: int):
    def ref_grad_func(
        value: torch.Tensor,
        dy: torch.Tensor,
        scale: torch.Tensor,
        offset: torch.Tensor,
    ):
        shape = [1 if axis != c else -1 for axis in range(t.ndim)]
        scale = scale.view(shape)
        offset = offset.view(shape)

        qt = ppq_tensor_round(value / scale, policy=RoundingPolicy.ROUND_HALF_EVEN)
        qt += offset
        clipped_qt = qt.clip(0, 255)

        dx = torch.where(clipped_qt != qt, torch.zeros_like(dy), dy)
        ds = torch.where(
            clipped_qt == qt,
            (((qt - offset) * scale) - value) * dy / scale,
            torch.zeros_like(dy),
        )
        ds += torch.where(qt > 255, (255 - offset) * dy, torch.zeros_like(dy))
        ds += torch.where(qt < 0, (0 - offset) * dy, torch.zeros_like(dy))
        ds = ds.transpose(0, c).flatten(1).sum(dim=-1) / math.sqrt(value.numel() * 255)
        return dx, ds

    with ENABLE_CUDA_KERNEL():
        nc = size[c]
        t = torch.rand(size=size).cuda() * 32
        s = torch.rand(size=[nc]).cuda()
        o = torch.randint(low=0, high=255, size=[nc]).float().cuda()
        dy = torch.rand_like(t)

        cuda_grad_x, cuda_grad_s = CUDA.LinearQuantize_C_B(
            t, s, o, dy, 0, 255, c, RoundingPolicy.ROUND_HALF_EVEN.value
        )

        ref_grad_x, ref_grad_s = ref_grad_func(t, dy, s, o)
        diff = ref_grad_x.flatten() - cuda_grad_x.flatten()
        assert diff.abs().max() == 0

        snr = torch_snr_error(
            cuda_grad_s.reshape([1, -1]), ref_grad_s.reshape([1, -1])
        ).item()
        if snr > 1e-5:
            assert cuda_grad_s.item() <= 1


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
@pytest.mark.parametrize("func", [CUDA.Histogram_T, CUDA.Histogram_Asymmetric_T])
def test_histogram(func):
    with ENABLE_CUDA_KERNEL():
        t = torch.rand(size=[128, 3, 224, 224]).to("cuda")
        s = 0.01

        # torch hist
        hist = torch.histc(torch.abs(t), bins=50, min=0, max=0.5)

        cuda_hist = torch.zeros(size=[50]).to("cuda").int()
        if func is CUDA.Histogram_T:
            cuda_hist = func(t, cuda_hist, s)
        elif func is CUDA.Histogram_Asymmetric_T:
            cuda_hist = func(0, 0.5, t, cuda_hist, clip_outliers=True)
        assert torch.abs(hist - cuda_hist).max().item() < 100


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
def test_quantile():
    percentile = 0.999
    with ENABLE_CUDA_KERNEL():
        t = torch.randn(128, 3, 224, 224).to("cuda")
        numel = t.numel()
        min_idx, max_idx = int(numel * (1 - percentile)), int(numel * percentile)
        # torch.kthvalue needs index from 1 to numel ...
        min_idx = max(0, min_idx) + 1
        max_idx = min(max_idx, numel - 1) + 1
        _min = torch.kthvalue(t.flatten(), k=min_idx, dim=0)[0].view(1, -1)
        _max = torch.kthvalue(t.flatten(), k=max_idx, dim=0)[0].view(1, -1)
        ref_stat = torch.cat([_max, _min], dim=-1)

        cuda_stat = CUDA.Quantile(t, percentile).view(1, -1)
        assert torch.abs(ref_stat - cuda_stat).max() == 0


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
def test_compute_mse_loss():
    def ref_mse_loss(histogram: list, start: int, step: int, end: int):
        num_of_elements = sum(histogram)
        loss = 0
        for idx, hist in enumerate(histogram):
            if idx < start:
                # 如果所选的 bin 已经超出了起点，那从 bin 的中心到起点的距离即
                # ((idx 到 起点的距离) + 0.5)
                # 注意 hist 统计时是直接取 floor 的，因此会在这里额外 - 1
                error = (start - idx - 1) + 0.5
            elif idx > end:
                # 注意 hist 统计时是直接取 floor 的
                error = (idx - end) + 0.5
            else:
                # 分别计算左右两侧的 err
                l_idx = (idx - start) % step
                r_idx = step - l_idx - 1
                if l_idx == r_idx:
                    error = l_idx + 0.25
                else:
                    l_err = l_idx + 0.5
                    r_err = r_idx + 0.5
                    error = min(l_err, r_err)
            loss += (hist * error * error) / num_of_elements
        return loss

    with ENABLE_CUDA_KERNEL():
        histograms = torch.randint(0, 10000, size=[50]).tolist()
        start, step, end = 10, 2, 20

        ref_loss = ref_mse_loss(histograms, start, step, end)
        cpu_loss = CUDA.compute_mse_loss(histograms, start, step, end)
        assert abs(ref_loss - cpu_loss) < 1e-4


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available.")
@pytest.mark.parametrize(
    "func",
    [
        CUDA.Histogram_C,
        CUDA.RoundingLoss_LC,
        CUDA.RoundingLoss_LC_B,
        CUDA.RoundingLoss_LT,
        CUDA.RoundingLoss_LT_B,
        CUDA.FloatingQuantize_C,
        CUDA.FloatingQuantize_C_B,
        CUDA.FloatingQuantize_T,
        CUDA.FloatingQuantize_T_B,
    ],
)
@pytest.mark.parametrize(
    ("size", "c"),
    [
        ([1, 1, 1, 1], 1),
        ([1, 1, 1, 1], 1),
        ([5, 12, 13, 4], 1),
        ([1, 7, 15, 41], 1),
        ([50, 120, 130, 4], 1),
        ([12, 74, 15, 411], 1),
        ([50, 7, 130, 1], 1),
        ([12, 4, 15, 3], 1),
        ([5011, 7, 7, 1], 0),
        ([122552, 1, 10, 4], 0),
        ([10, 10, 124, 47], 3),
        ([19, 42, 150, 3], 3),
    ],
)
def test_fuzzy_func_call(func, size: List[int], c: int):

    def args_from_signature(signature):
        signature = inspect.signature(func)
        nc = size[c]
        for k, v in signature.parameters.items():
            if k in ("tensor", "dy") and v.annotation == torch.Tensor:
                yield k, torch.randn(size).cuda() * 32
            elif k in ("histogram") and v.annotation == torch.Tensor:
                yield k, torch.randint(0, 256, size=[nc]).to(torch.int32).cuda()
            elif v.annotation == torch.Tensor:
                yield k, torch.randn([nc]).cuda()
            elif k == "channel_axis":
                yield k, c
            elif v.default is signature.empty and issubclass(v.annotation, Number):
                yield k, v.annotation()

    with ENABLE_CUDA_KERNEL():
        kw = {k: v for k, v in args_from_signature(func)}
        func(**kw)



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

"""

from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Collection, Iterable, Sequence
from typing import Any, Dict, Optional, Type

import torch

from mppq.data import DataType
from mppq.defs import empty_ppq_cache
from mppq.executor.base import BaseGraphExecutor, GraphInput
from mppq.executor.torch import TorchExecutor
from mppq.ir.base.command import QuantizeOperationCommand
from mppq.ir.base.graph import BaseGraph, Operation
from mppq.ir.base.quantize import QuantableOperation, QuantableVariable
from mppq.ir.deploy import QuantableGraph
from mppq.ir.morph import GraphReplacer
from mppq.logger import error, info, warning
from mppq.quant import OperationQuantizationConfig, QuantizationStates, TargetPrecision
from mppq.quantization.optim.base import (
    OPTIM_ALGORITHMS,
    QuantizationOptimizationPass,
    QuantizationOptimizationPipeline,
)
from mppq.register import Registry


class BaseQuantizer(metaclass=ABCMeta):
    r"""量化起的基类。提供了一个基本的量化框架。

    需要实现以下接口：

    1. init_quantize_config
       实现算子的量化配置的初始化

    2. quant_operation_types (property)
       所支持的量化算子类型

    可选择实现以下接口：

    1. default_prequant_pipeline (property)
       当未指定时，默认的量化前算法

    2. default_quant_pipeline (property)
       当未指定时，默认的量化算法
    """

    def __init__(self, graph: BaseGraph, verbose: bool = True) -> None:
        if not isinstance(graph, BaseGraph):
            raise TypeError(
                "To initialize a Quantizer, a BaseGraph instance is needed. "
                f"While {type(graph)} was given."
            )
        self._verbose = verbose
        self._graph = graph
        self._processor = QuantableGraph(GraphReplacer(self._graph))

    @property
    def graph(self) -> BaseGraph:
        return self._graph

    @empty_ppq_cache
    def quantize(
        self,
        example_inputs: Optional[GraphInput] = None,
        calib_dataloader: Optional[Iterable[Any]] = None,
        executor: Optional[BaseGraphExecutor] = None,
        collate_fn: Optional[Callable[[Any], Any]] = None,
        calib_steps: int = 32,
        prequant_settings: Optional[
            Sequence[dict | QuantizationOptimizationPass]
        ] = None,
        quant_settings: Optional[Sequence[dict | QuantizationOptimizationPass]] = None,
    ) -> BaseGraph:
        r"""Quantize the graph."""

        if executor is None:
            executor = TorchExecutor(self._graph)
        if example_inputs is None:
            new_example_inputs: Dict[str, torch.Tensor] = {}
            for k, v in self._graph.inputs.items():
                dtype = DataType.to_torch(v.dtype)
                shape = [int(i) for i in v.shape]
                new_example_inputs[k] = torch.zeros(shape, dtype=dtype)
            example_inputs = new_example_inputs  # type: ignore
        if calib_dataloader is None:
            calib_dataloader = [example_inputs]
        assert example_inputs is not None
        assert isinstance(executor, BaseGraphExecutor)

        # step - 1, prequant pipeline:
        # prequant pipeline will change your network structure and float value.
        if prequant_settings is None:
            prequant_pipeline = self.default_prequant_pipeline
        else:
            prequant_pipeline = self._build_pipeline(prequant_settings)
        prequant_pipeline.optimize(
            graph=self._graph,
            dataloader=calib_dataloader,
            executor=executor,
            verbose=self._verbose,
        )

        # step - 2, quantize all operations
        executor.load_graph(self._graph)
        executor.tracing_operation_meta(inputs=example_inputs)

        for op_name, operation in self._graph.operations.items():
            if operation.precision == TargetPrecision.UNSPECIFIED:
                error(f"no precision info for node {op_name}")
                raise ValueError("have you call a correct dispatcher?")
            self._wrap_to_quantized_operation(op_name)

        # quantize operation will modify network structure
        # it is necessary calling self._executor before further execution
        # step - 3, calling graph optimization pipeline
        executor.load_graph(self._graph)
        if quant_settings is None:
            quant_pipeline = self.default_quant_pipeline
        else:
            quant_pipeline = self._build_pipeline(quant_settings)

        quant_pipeline.optimize(
            graph=self._graph,
            dataloader=calib_dataloader,
            executor=executor,
            collate_fn=collate_fn,
            calib_steps=calib_steps,
            verbose=self._verbose,
        )

        if self._verbose:
            info(f"{self}")
            info("Network Quantization Finished.")
        return self._graph

    def _wrap_to_quantized_operation(self, op_name: str) -> Operation:
        if op_name not in self._graph.operations:
            raise KeyError(
                f"Can not find op {op_name} in your graph, check operation name again."
            )
        converting_operation = self._graph.operations[op_name]
        if isinstance(converting_operation, QuantableOperation):
            warning(f"Operation {op_name} has been quantized.")
            return converting_operation

        precision = converting_operation.precision
        if precision in {TargetPrecision.FP32, TargetPrecision.SOI}:
            return self._graph.operations[op_name]

        if (
            precision == TargetPrecision.UNSPECIFIED
            and converting_operation.type not in self.quant_operation_types
        ):
            return self._graph.operations[op_name]

        # create quantize config and convert operation.
        self._processor(
            QuantizeOperationCommand(
                op_name=op_name,
                target_precision=precision,
                config=self.init_quantize_config(operation=converting_operation),
            )
        )
        return self._graph.operations[op_name]

    def _build_pipeline(
        self, configurations: Sequence[dict | QuantizationOptimizationPass]
    ):
        r"""Build a pipeline of optimization algorithms."""
        list_of_passes = []
        for optim in configurations:
            if isinstance(optim, QuantizationOptimizationPass):
                list_of_passes.append(optim)
                continue
            assert isinstance(optim, dict) and "type" in optim
            algo = OPTIM_ALGORITHMS[optim.pop("type")]
            list_of_passes.append(algo(**optim))
        return QuantizationOptimizationPipeline(list_of_passes)

    @abstractmethod
    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        r"""Return a query to the operation how it should be quantized."""
        raise NotImplementedError

    @property
    def quant_operation_types(self) -> Collection[str]:
        r"""Return a collection of name that operations should be quantized."""
        return set()  # all operations are quantized by default.

    @property
    def default_prequant_pipeline(self) -> QuantizationOptimizationPipeline:
        r"""A simplified API to return a default quantization pipeline."""
        return QuantizationOptimizationPipeline([])

    @property
    def default_quant_pipeline(self) -> QuantizationOptimizationPipeline:
        r"""A simplified API to return a default quantization pipeline."""
        raise NotImplementedError

    def __repr__(self) -> str:
        debug_str = ""
        # stats:
        quant_ops = [
            op
            for op in self._graph.operations.values()
            if isinstance(op, QuantableOperation)
        ]
        quant_vars = [
            var
            for var in self._graph.variables.values()
            if isinstance(var, QuantableVariable)
        ]
        quant_cfgs = []

        config_states_cnt = {state: 0 for state in QuantizationStates}
        for op in quant_ops:
            for cfg, _ in op.config_with_variable:
                config_states_cnt[cfg.state] += 1
                quant_cfgs.append(cfg)

        debug_str += "--------- Network Snapshot ---------\n"
        debug_str += f"Num of Op:                    [{len(self._graph.operations)}]\n"
        debug_str += f"Num of Quantized Op:          [{len(quant_ops)}]\n"
        debug_str += f"Num of Variable:              [{len(self._graph.variables)}]\n"
        debug_str += f"Num of Quantized Var:         [{len(quant_vars)}]\n"
        debug_str += "------- Quantization Snapshot ------\n"
        debug_str += f"Num of Quant Config:          [{len(quant_cfgs)}]\n"
        for state, cnt in config_states_cnt.items():
            if cnt <= 0:
                continue
            padding = max(28 - len(state.name), 0)
            debug_str += f"{state.name}: [{cnt}]{'.':<{padding}}"
        return debug_str


QUANTIZER: Registry[Type[BaseQuantizer]] = Registry("QUANTIZER")



from typing import Collection, Dict

from mppq.dispatcher.base import DISPATCHER_TABLE, GraphDispatcher
from mppq.ir.base.graph import BaseGraph
from mppq.quant import TargetPrecision


@DISPATCHER_TABLE.register("allin")
class AllinDispatcher(GraphDispatcher):
    """Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.
    ATTENTION: this dispatcher will enable all ops in quant_types to quant_precision.
    """

    def __init__(self, graph: BaseGraph) -> None:
        self.graph = graph

    def dispatch(
        self, quant_types: Collection[str], quant_precision: TargetPrecision, **kwargs
    ) -> Dict[str, TargetPrecision]:
        """We assume all ops in origin model can be quant. This is suitable for some
        npu platform.

        Args:
            graph (BaseGraph): graph object which going to be dispatched by this
                dispatcher.
            quant_types(Set[str]): all quantable types for given platforms.

        Returns:
            Dict[str, TargetPlatform]: [description]
        """
        graph = self.graph

        dispatching_table: Dict[str, TargetPrecision] = {}
        for op in graph.operations.values():
            if op.type in quant_types:
                dispatching_table[op.name] = quant_precision
            else:
                dispatching_table[op.name] = TargetPrecision.FP32

        return dispatching_table



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

Default socket definition for common onnx operators.
"""

# pylint: disable=invalid-name

from mppq.ir.base.opdef import DEFAULT_SOCKET_TABLE, Operation, OpSocket, VLink
from mppq.quant import TargetPrecision


def Reshape(op: Operation) -> OpSocket:
    """
    From Opset 5 - 13:

    Inputs
        data (differentiable) : T
            An input tensor.

        shape (non-differentiable) : tensor(int64)
            Specified shape for output.

    Outputs
        reshaped (differentiable) : T
            Reshaped data.
    """

    return OpSocket(
        op=op,
        in_plat=[TargetPrecision.UNSPECIFIED, TargetPrecision.SOI],
        links=[VLink(in_idx=0, out_idx=0)],
    )


def Pad(op: Operation) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (2 - 3)
        data (differentiable) : T
            Input tensor.

        pads (non-differentiable) : tensor(int64)
            Tensor of integers indicating the number of padding elements to add or remove
            (if negative) at the beginning and end of each axis. For 2D input tensor, it is the number of pixels.
            `pads` should be a 1D tensor of shape [2 * input_rank].
            `pads` format should be: [x1_begin, x2_begin,...,x1_end, x2_end,...],
            where xi_begin is the number of pad values added at the beginning of axis `i` and xi_end,
            the number of pad values added at the end of axis `i`.

        constant_value (optional, non-differentiable) : T
            (Optional) A scalar value to be used if the mode chosen is `constant`
            (by default it is 0, empty string or False).

    Outputs
        output (differentiable) : T
            Tensor after padding.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.SOI, TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Gather(op: Operation) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : Tind
            Tensor of int32/int64 indices, of any rank q.
            All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

    Outputs
        output (differentiable) : T
            Tensor of rank q + (r - 1).
    """

    return OpSocket(
        op=op,
        in_plat=[TargetPrecision.UNSPECIFIED, TargetPrecision.FP32],
        links=[VLink(in_idx=0, out_idx=0)],
    )


def Resize(op: Operation) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs (1 - 4)
        X (differentiable) : T1
            N-D tensor

        roi (optional, non-differentiable) : T2
            1-D tensor given as [start1, ..., startN, end1, ..., endN],
            where N is the rank of X. The RoIs' coordinates are normalized in the coordinate system of the input image.
            It only takes effect when coordinate_transformation_mode is "tf_crop_and_resize"

        scales (optional, non-differentiable) : tensor(float)
            The scale array along each dimension. It takes value greater than 0.
            If it's less than 1, it's sampling down, otherwise, it's upsampling.
            The number of elements of 'scales' should be the same as the rank of input 'X'.
            One of 'scales' and 'sizes' MUST be specified and it is an error if both are specified.
            If 'sizes' is needed, the user can use an empty string as the name of 'scales' in this operator's input list.

        sizes (optional, non-differentiable) : tensor(int64)
            The size of the output tensor. The number of elements of 'sizes' should be the same as the rank of input 'X'.
            Only one of 'scales' and 'sizes' can be specified.

    Outputs
        Y (differentiable) : T1
            N-D tensor after resizing
    """

    in_plat = [
        TargetPrecision.UNSPECIFIED,
        TargetPrecision.SOI,
        TargetPrecision.SOI,
        TargetPrecision.SOI,
    ]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Split(op: Operation) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (1 - 2)
        input (differentiable) : T
            The tensor to split

        split (optional, non-differentiable) : tensor(int64)
            Optional length of each output. Values should be >= 0.Sum of the values
            must be equal to the dim value at 'axis' specified.

    Outputs (1 - ∞)
        outputs (variadic, differentiable) : T
            One or more outputs forming list of tensors after splitting
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Topk(op: Operation) -> OpSocket:
    """
    From Opset 1 - 11:

    Inputs
        X (differentiable) : T
            Tensor of shape [a_1, a_2, ..., a_n, r]

        K (non-differentiable) : tensor(int64)
            A 1-D tensor containing a single positive value corresponding to the number of top elements to retrieve

    Outputs
        Values (differentiable) : T
            Tensor of shape [a_1, a_2, ..., a_{axis-1}, k, a_{axis+1}, ... a_n]
            containing top K values from the input tensor

        Indices (non-differentiable) : I
            Tensor of shape [a_1, a_2, ..., a_{axis-1}, k, a_{axis+1}, ... a_n]
            containing the corresponding input tensor indices for the top K values.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.SOI]
    out_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.FP32]
    return OpSocket(
        op=op,
        in_plat=in_plat[: op.num_of_input],
        out_plat=out_plat,
        links=[VLink(in_idx=0, out_idx=0)],
    )


def Tile(op: Operation) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs
        input (differentiable) : T
            Input tensor of any shape.

    repeats (non-differentiable) : T1
        1D int64 tensor of the same length as input's dimension number,
        includes numbers of repeated copies along input's dimensions.

    Outputs
        output (differentiable) : T
            Output tensor of the same dimensions and type as tensor input.
            output_dim[i] = input_dim[i] * repeats[i]
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.SOI, TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Expand(op: Operation) -> OpSocket:
    """
    From Opset 8 - 13:

    Inputs
        input (differentiable) : T

    Input tensor
        shape (non-differentiable) : tensor(int64)
            A 1-D tensor indicates the shape you want to expand to, following the broadcast rule

    Outputs
        output (differentiable) : T
            Output tensor
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def RoiAlign(op: Operation) -> OpSocket:
    """
    From Opset 10 - 16:

    Inputs
        X : T1
            Input data tensor from the previous operator; 4-D feature map of shape (N, C, H, W),
            where N is the batch size, C is the number of channels,
            and H and W are the height and the width of the data.

        rois : T1
            RoIs (Regions of Interest) to pool over; rois is 2-D input of shape (num_rois, 4)
            given as [[x1, y1, x2, y2], ...]. The RoIs' coordinates are in the coordinate system of the input image.
            Each coordinate set has a 1:1 correspondence with the 'batch_indices' input.

        batch_indices : T2
            1-D tensor of shape (num_rois,) with each element denoting the index of the corresponding image in the batch.

    Outputs
        Y : T1
            RoI pooled output, 4-D tensor of shape (num_rois, C, output_height, output_width).
            The r-th batch element Y[r-1] is a pooled feature map corresponding to the r-th RoI X[r-1].
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.FP32, TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def GridSampler(op: Operation) -> OpSocket:
    """
    From MMCV
    """
    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.FP32]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Clip(op: Operation) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (1 - 3)
        input (differentiable) : T
            Input tensor whose elements to be clipped

        min (optional, non-differentiable) : T
            Minimum value, under which element is replaced by min.
            It must be a scalar(tensor of empty shape).

        max (optional, non-differentiable) : T
            Maximum value, above which element is replaced by max.
            It must be a scalar(tensor of empty shape).

    Outputs
        output (differentiable) : T
            Output tensor with clipped input elements
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.FP32, TargetPrecision.FP32]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def ConstantOfShape(op: Operation) -> OpSocket:
    """
    From Opset 9 - 9:

    Inputs
        input : T1
            1D tensor. The shape of the expected output tensor.
            If empty tensor is given, the output would be a scalar. All values must be >= 0.

    Outputs
        output : T2
            Output tensor of shape specified by 'input'.If attribute 'value' is specified,
            the value and datatype of the output tensor is taken from 'value'.If attribute 'value' is not specified,
            the value in the output defaults to 0, and the datatype defaults to float32.
    """

    in_plat = [TargetPrecision.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[])


def GatherElements(op: Operation) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : Tind
            Tensor of int32/int64 indices, with the same rank r as the input.
            All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

    Outputs
        output (differentiable) : T
            Tensor of the same shape as indices.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.FP32]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def GatherND(op: Operation) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : tensor(int64)
            Tensor of rank q >= 1. All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

    Outputs
        output (differentiable) : T
            Tensor of rank q + r - indices_shape[-1] - 1.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.FP32]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def NonMaxSuppression(op: Operation) -> OpSocket:
    """
    From Opset 10 - 13:

    Inputs (2 - 5)
        boxes : tensor(float)
            An input tensor with shape [num_batches, spatial_dimension, 4].
            The single box data format is indicated by center_point_box.

        scores : tensor(float)
            An input tensor with shape [num_batches, num_classes, spatial_dimension]

        max_output_boxes_per_class (optional) : tensor(int64)
            Integer representing the maximum number of boxes to be selected per batch per class.
            It is a scalar. Default to 0, which means no output.

        iou_threshold (optional) : tensor(float)
            Float representing the threshold for deciding whether boxes overlap too much with respect to IOU.
            It is scalar. Value range [0, 1]. Default to 0.

        score_threshold (optional) : tensor(float)
            Float representing the threshold for deciding when to remove boxes based on score. It is a scalar.

    Outputs
        selected_indices : tensor(int64)
            selected indices from the boxes tensor. [num_selected_indices, 3],
            the selected index format is [batch_index, class_index, box_index].
    """

    in_plat = [
        TargetPrecision.UNSPECIFIED,
        TargetPrecision.FP32,
        TargetPrecision.SOI,
        TargetPrecision.SOI,
        TargetPrecision.SOI,
    ]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[])


def NonZero(op: Operation) -> OpSocket:
    """
    From Opset 9 - 13:

    Inputs
        X (non-differentiable) : T
            input

    Outputs
        Y (non-differentiable) : tensor(int64)
            output
    """

    in_plat = [TargetPrecision.UNSPECIFIED]
    out_plat = [TargetPrecision.FP32]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[]
    )


def Range(op: Operation) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs
        start : T
            Scalar. First entry for the range of output values.

        limit : T
            Scalar. Exclusive upper limit for the range of output values.

        delta : T
            Scalar. Value to step by.

    Outputs
        output : T
            A 1-D tensor with same type as the inputs containing generated range of values.
    """

    in_plat = [TargetPrecision.SOI, TargetPrecision.SOI, TargetPrecision.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[])


def ScatterElements(op: Operation) -> OpSocket:
    """
    From Opset 11 - 16:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : Tind
            Tensor of int32/int64 indices, of r >= 1 (same rank as input).
            All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

        updates (differentiable) : T
            Tensor of rank r >=1 (same rank and shape as indices)

    Outputs
        output (differentiable) : T
            Tensor of rank r >= 1 (same rank as input).
    """

    in_plat = [
        TargetPrecision.UNSPECIFIED,
        TargetPrecision.FP32,
        TargetPrecision.UNSPECIFIED,
    ]
    return OpSocket(
        op=op,
        in_plat=in_plat[: op.num_of_input],
        links=[VLink(in_idx=0, out_idx=0), VLink(in_idx=2, out_idx=0)],
    )


def ScatterND(op: Operation) -> OpSocket:
    """
    From Opset 11 - 16:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : tensor(int64)
            Tensor of rank q >= 1.

        updates (differentiable) : T
            Tensor of rank q + r - indices_shape[-1] - 1.

    Outputs
        output (differentiable) : T
            Tensor of rank r >= 1.
    """

    in_plat = [
        TargetPrecision.UNSPECIFIED,
        TargetPrecision.FP32,
        TargetPrecision.UNSPECIFIED,
    ]
    return OpSocket(
        op=op,
        in_plat=in_plat[: op.num_of_input],
        links=[VLink(in_idx=0, out_idx=0), VLink(in_idx=2, out_idx=0)],
    )


def Shape(op: Operation) -> OpSocket:
    """
    From Opset 1 - 16:

    Inputs
        data (non-differentiable) : T
            An input tensor.

    Outputs
        shape (non-differentiable) : T1
            Shape of the input tensor
    """

    in_plat = [TargetPrecision.UNSPECIFIED]
    out_plat = [TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[]
    )


def Slice(op: Operation) -> OpSocket:
    """
    From Opset 1 - 20:

    Inputs (3 - 5)
        data : T
            Tensor of data to extract slices from.

        starts : Tind
            1-D tensor of starting indices of corresponding axis in `axes`

        ends : Tind
            1-D tensor of ending indices (exclusive) of corresponding axis in `axes`

        axes (optional) : Tind
            1-D tensor of axes that `starts` and `ends` apply to.
            Negative value means counting dimensions from the back. Accepted range is [-r, r-1] where r = rank(data).

        steps (optional) : Tind
            1-D tensor of slice step of corresponding axis in `axes`.
            Negative value means slicing backward. 'steps' cannot be 0. Defaults to 1.

    Outputs
        output : T
            Sliced data tensor.
    """

    in_plat = [
        TargetPrecision.UNSPECIFIED,
        TargetPrecision.SOI,
        TargetPrecision.SOI,
        TargetPrecision.SOI,
        TargetPrecision.SOI,
    ]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Squeeze(op: Operation) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (1 - 2)
        data (differentiable) : T
            Tensors with at least max(dims) dimensions.

        axes (optional, non-differentiable) : tensor(int64)
            List of integers indicating the dimensions to squeeze.
            Negative value means counting dimensions from the back.
            Accepted range is [-r, r-1] where r = rank(data).

    Outputs
        squeezed (differentiable) : T
            Reshaped tensor with same data as input.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Unsqueeze(op: Operation) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs
        data (differentiable) : T
            Original tensor

        axes (non-differentiable) : tensor(int64)
            List of integers indicating the dimensions to be inserted.
            Negative value means counting dimensions from the back.
            Accepted range is [-r, r-1] where r = rank(expanded).

    Outputs
        expanded (differentiable) : T
            Reshaped tensor with same data as input.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Where(op: Operation) -> OpSocket:
    """
    From Opset 9 - 16:

    Inputs
        condition (non-differentiable) : B
            When True (nonzero), yield X, otherwise yield Y

        X (differentiable) : T
            values selected at indices where condition is True

        Y (differentiable) : T
            values selected at indices where condition is False

    Outputs
        output (differentiable) : T
            Tensor of shape equal to the broadcasted shape of condition, X, and Y.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.FP32, TargetPrecision.FP32]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)]
    )


def Logical(op: Operation) -> OpSocket:
    """
    From Opset 1 - 16:

    Inputs
        A (non-differentiable) : T
            First input operand for the logical operator.

        B (non-differentiable) : T
            Second input operand for the logical operator.

    Outputs
        C (non-differentiable) : T1
            Result tensor.
    """

    in_plat = [TargetPrecision.UNSPECIFIED, TargetPrecision.UNSPECIFIED]
    out_plat = [TargetPrecision.FP32]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[]
    )


def Onehot(op: Operation) -> OpSocket:
    """
    Inputs
        indices (non-differentiable) : T1

        depth (non-differentiable) : T2

        values (non-differentiable) : T3

    Outputs
        output (non-differentiable) : T3

    Args:
        op (Operation): _description_

    Returns:
        OpSocket: _description_
    """

    in_plat = [TargetPrecision.SOI, TargetPrecision.SOI, TargetPrecision.SOI]
    out_plat = [TargetPrecision.SOI]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[]
    )


DEFAULT_SOCKET_TABLE.update(
    Clip=Clip,
    ConstantOfShape=ConstantOfShape,
    Equal=Logical,
    Expand=Expand,
    Gather=Gather,
    GatherElements=GatherElements,
    GatherND=GatherND,
    grid_sampler=GridSampler,
    Greater=Logical,
    Less=Logical,
    NonMaxSuppression=NonMaxSuppression,
    NonZero=NonZero,
    Pad=Pad,
    Range=Range,
    Reshape=Reshape,
    Resize=Resize,
    ScatterElements=ScatterElements,
    ScatterND=ScatterND,
    Shape=Shape,
    Slice=Slice,
    Split=Split,
    Squeeze=Squeeze,
    Tile=Tile,
    TopK=Topk,
    Unsqueeze=Unsqueeze,
    Where=Where,
    RoiAlign=RoiAlign,
    MMCVRoiAlign=RoiAlign,
    OneHot=Onehot,
    GreaterOrEqual=Logical,
    LessOrEqual=Logical,
    Xor=Logical,
    Or=Logical,
    And=Logical,
)



"""PPQ Core Data Structure Abstraction PPQ 核心数据结构抽象.

You are not allowed to modify this 请勿修改此文件
"""

from enum import Enum
from numbers import Number
from typing import Any, Literal, Optional, Sequence, overload

import numpy as np
import torch
from numpy import dtype as np_type
from onnx import TensorProto
from torch import dtype as torch_type


class DataType(Enum):
    """
    DataType defines all PPQ internal data type and its enumeration value.
        ATTENTION: PPQ shares same data type enumeration value with Onnx.

    System maintainers and modifier are supposed to keep this corresponding.
    Cause OnnxExporter directly use this value to export PPQ graph towards Onnx.
    """

    INT4 = TensorProto.INT4
    UINT4 = TensorProto.UINT4
    INT8 = TensorProto.INT8
    UINT8 = TensorProto.UINT8
    INT16 = TensorProto.INT16
    UINT16 = TensorProto.UINT16
    INT32 = TensorProto.INT32
    UINT32 = TensorProto.UINT32
    INT64 = TensorProto.INT64
    UINT64 = TensorProto.UINT64

    FP8_E5M2 = TensorProto.FLOAT8E5M2
    BF16 = TensorProto.BFLOAT16
    FP16 = TensorProto.FLOAT16
    FP32 = TensorProto.FLOAT
    FP64 = TensorProto.DOUBLE

    BOOL = TensorProto.BOOL
    COMPLEX128 = TensorProto.COMPLEX128
    COMPLEX64 = TensorProto.COMPLEX64
    NONETYPE = TensorProto.UNDEFINED

    @classmethod
    def from_numpy(cls, dtype: np_type):
        """Construct DataType from numpy dtype."""
        numpy_converting_dict = {
            np_type("bool"): DataType.BOOL,
            np_type("uint8"): DataType.UINT8,
            np_type("int8"): DataType.INT8,
            np_type("int16"): DataType.INT16,
            np_type("int32"): DataType.INT32,
            np_type("int64"): DataType.INT64,
            np_type("float16"): DataType.FP16,
            np_type("float32"): DataType.FP32,
            np_type("float64"): DataType.FP64,
            np.bool: DataType.BOOL,
            np.uint8: DataType.UINT8,
            np.int8: DataType.INT8,
            np.int16: DataType.INT16,
            np.int32: DataType.INT32,
            np.int64: DataType.INT64,
            np.float16: DataType.FP16,
            np.float32: DataType.FP32,
            np.float64: DataType.FP64,
        }
        if dtype not in numpy_converting_dict:
            raise TypeError(
                f"Numpy type {dtype} is not included in ppq now. "
                "please contact with system developer."
            )
        else:
            return numpy_converting_dict[dtype]

    @classmethod
    def from_torch(cls, dtype: torch_type):
        """Construct DataType from torch dtype."""
        torch_converting_dict = {
            torch.bool: DataType.BOOL,
            torch.uint8: DataType.UINT8,
            torch.int8: DataType.INT8,
            torch.int16: DataType.INT16,
            torch.int32: DataType.INT32,
            torch.int64: DataType.INT64,
            torch.float16: DataType.FP16,
            torch.float32: DataType.FP32,
            torch.float64: DataType.FP64,
            torch.float8_e5m2: DataType.FP8_E5M2,
        }
        if dtype not in torch_converting_dict:
            raise TypeError(
                f"Torch dtype {dtype} is not included in ppq now. "
                "please contact with system developer."
            )
        else:
            return torch_converting_dict[dtype]

    @classmethod
    def to_numpy(cls, dtype) -> np_type:
        """Convert DataType to numpy dtype."""
        numpy_converting_dict = {
            DataType.BOOL: np_type("bool"),
            DataType.UINT8: np_type("uint8"),
            DataType.INT8: np_type("int8"),
            DataType.INT16: np_type("int16"),
            DataType.INT32: np_type("int32"),
            DataType.INT64: np_type("int64"),
            DataType.FP16: np_type("float16"),
            DataType.FP32: np_type("float32"),
            DataType.FP64: np_type("float64"),
        }
        assert isinstance(dtype, DataType)
        return numpy_converting_dict[dtype]

    @classmethod
    def to_torch(cls, dtype) -> torch_type:
        """Convert DataType to torch dtype."""
        torch_converting_dict = {
            DataType.BOOL: torch.bool,
            DataType.UINT8: torch.uint8,
            DataType.INT8: torch.int8,
            DataType.INT16: torch.int16,
            DataType.INT32: torch.int32,
            DataType.INT64: torch.int64,
            DataType.FP16: torch.float16,
            DataType.FP32: torch.float32,
            DataType.FP64: torch.float64,
            DataType.FP8_E5M2: torch.float8_e5m2,
        }
        assert isinstance(dtype, DataType)
        return torch_converting_dict[dtype]


@overload
def convert_any_to_python_primary_type(
    x: Any, accept_none: Literal[True] = True
) -> Number | str | bool | list | None:
    """Return contains None type"""


@overload
def convert_any_to_python_primary_type(
    x: Any, accept_none: Literal[False]
) -> Number | str | bool | list:
    """Return does not contain None type"""


def convert_any_to_python_primary_type(x: Any, accept_none: bool = True):
    """Try to convert an object to python POD."""

    if x is None and accept_none:
        return None
    if x is None and not accept_none:
        raise ValueError("Trying to convert an empty value.")
    if isinstance(x, (Number, str, bool)):
        return x
    elif isinstance(x, torch.Tensor):
        if x.numel() == 0 and accept_none:
            return None
        if x.numel() == 0 and not accept_none:
            raise ValueError("Trying to convert an empty value.")
        if str(x.device) != "cpu":
            x = x.cpu()
        if x.numel() == 1:
            return x.item()
        if x.numel() > 1:
            return x.tolist()
    elif isinstance(x, np.ndarray):
        if x.size == 0 and accept_none:
            return None
        if x.size == 0 and not accept_none:
            raise ValueError("Trying to convert an empty value.")
        if x.size == 1:
            return x.flatten().tolist()[0]
        if x.size > 1:
            return x.tolist()
    elif isinstance(x, Sequence):
        return list(x)
    else:
        raise TypeError(
            f"input value {x}({type(x)}) can not be converted as python primary type."
        )


@overload
def convert_any_to_numpy(
    x: Any, accept_none: Literal[True] = True
) -> None | np.ndarray:
    """Return contains None type"""


@overload
def convert_any_to_numpy(x: Any, accept_none: Literal[False]) -> np.ndarray:
    """Return does not contain None type"""


def convert_any_to_numpy(x: Any, accept_none: bool = True) -> None | np.ndarray:
    """Try to convert an object to numpy array."""
    if x is None and accept_none:
        return None
    if x is None and not accept_none:
        raise ValueError("Trying to convert an empty value.")
    if isinstance(x, np.ndarray):
        if x.size == 0 and accept_none:
            return None
        if x.size == 0 and not accept_none:
            raise ValueError("Trying to convert an empty value.")
        return x
    elif isinstance(x, torch.Tensor):
        return convert_any_to_numpy(x.cpu().numpy(), accept_none=accept_none)
    elif isinstance(x, Number):
        return np.array([x])
    elif isinstance(x, (Sequence)):
        return np.array(x)
    else:
        raise TypeError(
            f"input value {x}({type(x)}) can not be converted as numpy type."
        )


def convert_any_to_tensor(
    x: Any,
    dtype: Optional[torch.dtype] = None,
    device: str | torch.device | None = "cpu",
) -> torch.Tensor:
    """Try to convert an object to torch tensor."""

    if isinstance(x, int):
        if dtype is None:
            dtype = torch.int64
        return torch.tensor(x, dtype=dtype, device=device)
    elif isinstance(x, float):
        if dtype is None:
            dtype = torch.float32
        return torch.tensor(x, dtype=dtype, device=device)
    elif isinstance(x, torch.Tensor):
        return x.to(dtype=dtype, device=device)
    elif isinstance(x, np.ndarray):
        return torch.from_numpy(x).to(dtype=dtype, device=device)
    elif isinstance(x, list) or isinstance(x, tuple):
        if all([isinstance(element, int) for element in x]):
            if dtype is None:
                dtype = torch.int64
        return torch.tensor(x, dtype=dtype, device=device)
    else:
        raise TypeError(
            f"input value {x}({type(x)}) can not be converted as torch tensor."
        )



"""
Copyright mPPQ/PPQ 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

Extend PPQ with customized platforms
"""

from copy import deepcopy
from typing import Dict, Optional, Type

from mppq.dispatcher.base import DISPATCHER_TABLE, GraphDispatcher
from mppq.executor.base import OPERATION_FORWARD_TABLE
from mppq.executor.op.base import DEFAULT_BACKEND_TABLE, OperationForwardProtocol
from mppq.frontend import EXPORTER, PARSER
from mppq.ir.base.graph import GraphBuilder, GraphExporter
from mppq.logger import error, warning
from mppq.quantizer.base import QUANTIZER, BaseQuantizer

_PLATFORM_TO_PARSER_: Dict[int, str] = {}
_PLATFORM_TO_EXPORTER_: Dict[int, str] = {}
_PLATFORM_TO_DISPATCHER_: Dict[int, str] = {}
_PLATFORM_TO_QUANTIZER_: Dict[int, str] = {}


def _register_operation_handler(
    handler: OperationForwardProtocol, operation_type: str, platform: int
):
    if platform not in OPERATION_FORWARD_TABLE:
        raise ValueError(
            "Unknown Platform detected, Please check your platform setting."
        )
    OPERATION_FORWARD_TABLE[platform].update({operation_type: handler})


def register_operation(op_type: str, platform: Optional[int] = None):
    r"""Register a new operation to forward table.

    Function should accept at least 3 input parameters, return one or more tensor
    as results:

    .. code-block:: python

        def func(
            op: Operation,
            values: List[torch.Tensor],
            ctx: Optional[TorchBackendContext] = None,
            **kwargs,
        ) -> torch.Tensor | Tuple[torch.Tensor,...]:
            ...

    If there is already another operation handler for given operation_type,
        new handler will replace the old one without warning.

    Args:
        op_type (str): name of the operation type
        platform (TargetPlatform, optional): specify a platform to register, if
            platform is None, add to the default backend table. Defaults to None.
    """

    def _wrapper(func: OperationForwardProtocol) -> OperationForwardProtocol:
        if platform is None:
            for i in OPERATION_FORWARD_TABLE.keys():
                _register_operation_handler(func, op_type, i)
        else:
            _register_operation_handler(func, op_type, platform)

        return func

    return _wrapper


def register_platform(
    platform_id: int,
    dispatcher: Dict[str | None, Optional[Type[GraphDispatcher]]],
    quantizer: Dict[str | None, Optional[Type[BaseQuantizer]]],
    parsers: Optional[Dict[str | None, Type[GraphBuilder]]] = None,
    exporters: Optional[Dict[str | None, Type[GraphExporter]]] = None,
):
    r"""注册一个新的自定义平台。

    Args:
        platform_id (int): 自定义平台的 ID ，不可重复。
        dispatcher (Dict[str | None, Type[GraphDispatcher]]): 自定义平台的调度器。
            字典键值可选，作为调度器的命名，和调度器的类类型。
        quantizer (Dict[str | None, Type[BaseQuantizer]]): 自定义平台的量化器。
            字典键值可选，作为量化器的命名，和量化器的类类型。
        parsers (Optional[Dict[str | None, Type[GraphBuilder]]], optional): 自定义平台的图构建器。
            字典键值可选，作为图构建器的命名，和图构建器的类类型。 Defaults to None.
        exporters (Optional[Dict[str | None, Type[GraphExporter]]], optional): 自定义平台的图导出器。
            字典键值可选，作为图导出器的命名，和图导出器的类类型。 Defaults to None.
    """
    if platform_id in _PLATFORM_TO_DISPATCHER_:
        raise KeyError(f"Platform {platform_id} is already registered.")
    OPERATION_FORWARD_TABLE[platform_id] = deepcopy(DEFAULT_BACKEND_TABLE)

    if parsers is not None:
        for i, (name, parser_cls) in enumerate(parsers.items()):
            if (name or parser_cls) in PARSER:
                warning(f"Parser {name} is replaced by {parser_cls}")
            PARSER.register(name)(parser_cls)
            name = PARSER.query_name(parser_cls)
            if i > 0:
                warning(
                    f"more than 1 parser is registered to platform {platform_id}, "
                    f"by default the last one is used. Current parser is {name}"
                )
            _PLATFORM_TO_PARSER_[platform_id] = name
    else:
        _PLATFORM_TO_PARSER_[platform_id] = "onnx"

    if exporters is not None:
        for i, (name, exporter_cls) in enumerate(exporters.items()):
            if (name or exporter_cls) in EXPORTER:
                warning(f"Exporter {name} is replaced by {exporter_cls}")
            EXPORTER.register(name)(exporter_cls)
            name = EXPORTER.query_name(exporter_cls)
            if i > 0:
                warning(
                    f"more than 1 exporter is registered to platform {platform_id}, "
                    f"by default the last one is used. Current exporter is {name}"
                )
            _PLATFORM_TO_EXPORTER_[platform_id] = name
    else:
        _PLATFORM_TO_EXPORTER_[platform_id] = "onnx"

    for i, (name, dispatcher_cls) in enumerate(dispatcher.items()):
        assert name is not None or dispatcher_cls is not None
        if (name or dispatcher_cls) in DISPATCHER_TABLE and dispatcher_cls is not None:
            warning(f"Dispatcher {name} is replaced by {dispatcher_cls}")
        if dispatcher_cls is not None:
            DISPATCHER_TABLE.register(name)(dispatcher_cls)
            name = DISPATCHER_TABLE.query_name(dispatcher_cls)
        elif not name:
            error("Both name and dispatcher are None.")
            raise ValueError
        if i > 0:
            warning(
                f"more than 1 dispatcher is registered to platform {platform_id}, "
                f"by default the last one is used. Current dispatcher is {name}"
            )
        _PLATFORM_TO_DISPATCHER_[platform_id] = name

    _PLATFORM_TO_QUANTIZER_[platform_id] = "default"
    for i, (name, quantizer_cls) in enumerate(quantizer.items()):
        if (name or quantizer_cls) in QUANTIZER and quantizer_cls is not None:
            warning(f"Quantizer {name} is replaced by {quantizer_cls}")
        if quantizer_cls is not None:
            QUANTIZER.register(name)(quantizer_cls)
            name = QUANTIZER.query_name(quantizer_cls)
        elif not name:
            error("Both name and quantizer are None.")
            raise ValueError
        if i > 0:
            warning(
                f"more than 1 quantizer is registered to platform {platform_id}, "
                f"by default the last one is used. Current quantizer is {name}"
            )
        _PLATFORM_TO_QUANTIZER_[platform_id] = name



from mppq.quantization.optim.adaround import AdaroundPass
from mppq.quantization.optim.baking import ParameterBakingPass
from mppq.quantization.optim.base import QuantizationOptimizationPipeline
from mppq.quantization.optim.calibration import (
    IsotoneCalibrationPass,
    RuntimeCalibrationPass,
)
from mppq.quantization.optim.equalization import (
    ActivationEqualizationPass,
    ChannelwiseSplitPass,
    LayerwiseEqualizationPass,
)
from mppq.quantization.optim.morph import GRUSplitPass, HorizontalLayerSplitPass
from mppq.quantization.optim.parameters import (
    ParameterQuantizePass,
    PassiveParameterQuantizePass,
)
from mppq.quantization.optim.refine import (
    MishFusionPass,
    QuantAlignmentPass,
    QuantizeFusionPass,
    QuantizeSimplifyPass,
    SwishFusionPass,
)
from mppq.quantization.optim.ssd import SSDEqualizationPass
from mppq.quantization.optim.training import (
    BiasCorrectionPass,
    LearnedStepSizePass,
    RoundTuningPass,
)

__all__ = [
    "ParameterBakingPass",
    "QuantizationOptimizationPipeline",
    "IsotoneCalibrationPass",
    "RuntimeCalibrationPass",
    "ActivationEqualizationPass",
    "ChannelwiseSplitPass",
    "LayerwiseEqualizationPass",
    "AdaroundPass",
    "GRUSplitPass",
    "HorizontalLayerSplitPass",
    "ParameterQuantizePass",
    "PassiveParameterQuantizePass",
    "MishFusionPass",
    "QuantAlignmentPass",
    "QuantizeFusionPass",
    "QuantizeSimplifyPass",
    "SwishFusionPass",
    "SSDEqualizationPass",
    "BiasCorrectionPass",
    "LearnedStepSizePass",
    "RoundTuningPass",
]



from mppq.quantization.observer.base import CalibrationHook, OperationObserver
from mppq.quantization.observer.dbdc import TorchDbdcObserver
from mppq.quantization.observer.floating import ConstantObserver, DirectMSEObserver
from mppq.quantization.observer.hist import TorchHistObserver, TorchHistogramObserver
from mppq.quantization.observer.isotone import TorchIsotoneObserver
from mppq.quantization.observer.min_max import TorchMinMaxObserver
from mppq.quantization.observer.mse import TorchMSEObserver
from mppq.quantization.observer.percentile import TorchPercentileObserver

__all__ = [
    "CalibrationHook",
    "OperationObserver",
    "TorchDbdcObserver",
    "ConstantObserver",
    "DirectMSEObserver",
    "TorchHistObserver",
    "TorchHistogramObserver",
    "TorchIsotoneObserver",
    "TorchMinMaxObserver",
    "TorchMSEObserver",
    "TorchPercentileObserver",
]



from typing import Tuple

from mppq.common import OBSERVER_MIN_SCALE, OBSERVER_MIN_SCALE_MANUL_OVERRIDE
from mppq.defs import ppq_quant_param_computing_function
from mppq.logger import warning
from mppq.quant import QuantizationProperty, RoundingPolicy, TensorQuantizationConfig
from mppq.utils.round import ppq_numerical_round, ppq_round_to_power_of_2


@ppq_quant_param_computing_function
def minmax_to_scale_offset(
    min_val: float,
    max_val: float,
    config: TensorQuantizationConfig,
    scale_threshold: float = OBSERVER_MIN_SCALE,
) -> Tuple[float, float]:
    """
    Solve scale and offset with given min, max value.
    For Symmetrical Quantization, offset is set to 0.
    For ASymmetrical Quantization, offset is limited by
    [config.quant_min, config.quant_max].

    Scale is limited by [scale_threshold, +inf].

    Args:
        min_val (float): min value
        max_val (float): max value
        config (TensorQuantizationConfig): Corresponding TQC.
        scale_threshold (float, optional): minimum scale.

    Returns:
        Tuple[float, float]: Solved scale and offset.
    """
    if OBSERVER_MIN_SCALE_MANUL_OVERRIDE in config.detail:
        scale_threshold = config.detail[OBSERVER_MIN_SCALE_MANUL_OVERRIDE]

    scale, offset = 1, 0
    if min_val > 0:
        min_val = 0
    if max_val < 0:
        max_val = 0

    if config.policy.has_property(QuantizationProperty.ASYMMETRIC):
        val_range = float(max_val - min_val)
        scale = val_range / (config.quant_max - config.quant_min)
        if scale < scale_threshold:
            warning(
                "Numeric instability detected: "
                "ppq find there is a scale value < 1e-7, "
                "which probably cause numeric underflow in further computation."
            )
        scale = max(scale, scale_threshold)
        offset = ppq_numerical_round(-min_val / scale)
    elif config.policy.has_property(QuantizationProperty.SYMMETRIC):
        val_range = 2 * float(max(abs(max_val), abs(min_val)))
        scale = val_range / (config.quant_max - config.quant_min)
        if scale < scale_threshold:
            warning(
                "Numeric instability detected: "
                "ppq find there is a scale value < 1e-7, "
                "which probably cause numeric underflow in further computation."
            )
        scale = max(scale, scale_threshold)
        offset = 0
    else:
        raise TypeError(
            "Tensor Min Max Observer Excepts either ASYMMETRICAL or SYMMETRICAL "
            "quantization config."
        )
    if config.policy.has_property(QuantizationProperty.POWER_OF_2):
        scale = ppq_round_to_power_of_2(scale, policy=RoundingPolicy.ROUND_UP)
    return scale, offset
